# SPDX-License-Identifier: BSD-2-Clause

import functools
import warnings
from abc               import ABCMeta, abstractmethod
from collections       import OrderedDict
from collections.abc   import (
	Iterable, MutableMapping, MutableSequence, MutableSet
)
from enum              import Enum
from itertools         import chain
from typing            import Optional, Union

from ..util            import flatten, tracer, union
from ..util.decorators import final, deprecated
from ..util.units      import bits_for
from ._unused          import MustUse, UnusedMustUse

__all__ = (
	'AnyConst',
	'AnySeq',
	'Array',
	'ArrayProxy',
	'Assert',
	'Assign',
	'Assume',
	'Cat',
	'ClockSignal',
	'Const',
	'Cover',
	'Fell',
	'Initial',
	'Mux',
	'Operator',
	'Part',
	'Past',
	'Property',
	'Repl',
	'ResetSignal',
	'Rose',
	'Sample',
	'Shape',
	'ShapeCastable',
	'Signal',
	'SignalDict',
	'SignalKey',
	'SignalSet',
	'signed',
	'Slice',
	'Stable',
	'Statement',
	'Switch',
	'unsigned',
	'Value',
	'ValueCastable',
	'ValueDict',
	'ValueKey',
	'ValueSet',
)


class DUID:
	''' Deterministic Unique IDentifier. '''
	__next_uid = 0

	def __init__(self) -> None:
		self.duid = DUID.__next_uid
		DUID.__next_uid += 1


class ShapeCastable:
	'''
	Interface of user-defined objects that can be cast to :class:`Shape`s.

	An object deriving from :class:`ShapeCastable` is automatically converted to a :class:`Shape`
	when it is used in a context where a :class:`Shape` is expected. Such objects can contain
	a richer description of the shape than what is supported by the core Torii language, yet
	still be transparently used with it.

	'''

	def __init_subclass__(cls, **kwargs):
		if not hasattr(cls, 'as_shape'):
			raise TypeError(f'Class \'{cls.__name__}\' deriving from `ShapeCastable` must override the `as_shape` method')



class Shape:
	'''
	Bit width and signedness of a value.

	A ``Shape`` can be constructed using:

	  * explicit bit width and signedness;
	  * aliases :func:`signed` and :func:`unsigned`;
	  * casting from a variety of objects.

	A ``Shape`` can be cast from:

	  * an integer, where the integer specifies the bit width;
	  * a range, where the result is wide enough to represent any element of the range, and is
		signed if any element of the range is signed;
	  * an :class:`Enum` with all integer members or :class:`IntEnum`, where the result is wide
		enough to represent any member of the enumeration, and is signed if any member of
		the enumeration is signed.

	Parameters
	----------
	width : int
		The number of bits in the representation, including the sign bit (if any).
	signed : bool
		If ``False``, the value is unsigned. If ``True``, the value is signed two's complement.

	'''  # noqa: E101

	def __init__(self, width: int = 1, signed: bool = False) -> None:
		if not isinstance(width, int):
			raise TypeError(f'Width must be an integer, not {width!r}')
		if not signed and width < 0:
			raise ValueError(f'Width of an unsigned value must be zero or a positive integer, not {width}')
		if signed and width <= 0:
			raise ValueError(f'Width of a signed value must be a positive integer, not {width}')

		self.width = width
		self.signed = signed

	# This implements an algorithm for inferring shape from standard Python enumerations
	# for `Shape.cast()`.
	@staticmethod
	def _cast_plain_enum(obj):
		signed = False
		width  = 0
		for member in obj:
			try:
				member_shape = Const.cast(member.value).shape()
			except TypeError:
				raise TypeError(
					'Only enumerations whose members have constant-castable '
					'values can be used in Torii code'
				)
			if not signed and member_shape.signed:
				signed = True
				width  = max(width + 1, member_shape.width)
			elif signed and not member_shape.signed:
				width  = max(width, member_shape.width + 1)
			else:
				width  = max(width, member_shape.width)
		return Shape(width, signed)

	@staticmethod
	def cast(
		obj: Union['Shape', int, range, type, ShapeCastable], *,
		src_loc_at: int = 0
	) -> 'Shape':
		while True:
			if isinstance(obj, Shape):
				return obj
			elif isinstance(obj, ShapeCastable):
				new_obj = obj.as_shape()
			elif isinstance(obj, int):
				return Shape(obj)
			elif isinstance(obj, range):
				if len(obj) == 0:
					return Shape(0)
				signed = obj[0] < 0 or obj[-1] < 0
				width  = max(
					bits_for(obj[0], signed),
					bits_for(obj[-1], signed)
				)
				return Shape(width, signed)
			elif isinstance(obj, type) and issubclass(obj, Enum):
				return Shape._cast_plain_enum(obj)
			else:
				raise TypeError(f'Object {obj!r} cannot be converted to a Torii shape')
			if new_obj is obj:
				raise RecursionError(f'Shape-castable object {obj!r} casts to itself')
			obj = new_obj

	def __repr__(self) -> str:
		if self.signed:
			return f'signed({self.width})'
		else:
			return f'unsigned({self.width})'

	def __eq__(self, other: Union[tuple[int, bool], 'Shape']) -> bool:
		if not isinstance(other, Shape):
			try:
				other = self.__class__.cast(other)
			except TypeError as e:
				raise TypeError(f'Shapes may be compared with shape-castable objects, not {other!r}') from e
		return self.width == other.width and self.signed == other.signed


def unsigned(width: int) -> Shape:
	''' Shorthand for ``Shape(width, signed = False)``. '''
	return Shape(width, signed = False)


def signed(width: int) -> Shape:
	''' Shorthand for ``Shape(width, signed = True)``. '''
	return Shape(width, signed = True)


ValueCastType = Union['Value', int, Enum, 'ValueCastable']

class Value(metaclass = ABCMeta):
	@staticmethod
	def cast(obj: ValueCastType ) -> 'Value':
		'''
		Converts ``obj`` to an Torii value.

		Booleans and integers are wrapped into a :class:`Const`. Enumerations whose members are
		all integers are converted to a :class:`Const` with a shape that fits every member.
		:class:`ValueCastable` objects are recursively cast to an Torii value.

		'''

		while True:
			if isinstance(obj, Value):
				return obj
			elif isinstance(obj, ValueCastable):
				new_obj = obj.as_value()
			elif isinstance(obj, Enum):
				return Const(obj.value, Shape.cast(type(obj)))
			elif isinstance(obj, int):
				return Const(obj)
			else:
				raise TypeError(f'Object {obj!r} cannot be converted to a Torii value')
			if new_obj is obj:
				raise RecursionError(f'Value-castable object {obj!r} casts to itself')
			obj = new_obj

	def __init__(self, *, src_loc_at: int = 0) -> None:
		super().__init__()
		self.src_loc = tracer.get_src_loc(1 + src_loc_at)

	def __bool__(self) -> None:
		raise TypeError('Attempted to convert Torii value to Python boolean')

	def __pos__(self) -> 'Value':
		return self

	def __invert__(self) -> 'Operator':
		return Operator('~', [self])

	def __neg__(self) -> 'Operator':
		return Operator('-', [self])

	def __add__(self, other: ValueCastType) -> 'Operator':
		return Operator('+', [self, other])

	def __radd__(self, other: ValueCastType) -> 'Operator':
		return Operator('+', [other, self])

	def __sub__(self, other: ValueCastType) -> 'Operator':
		return Operator('-', [self, other])

	def __rsub__(self, other: ValueCastType) -> 'Operator':
		return Operator('-', [other, self])


	def __mul__(self, other: ValueCastType) -> 'Operator':
		return Operator('*', [self, other])

	def __rmul__(self, other: ValueCastType) -> 'Operator':
		return Operator('*', [other, self])

	def __mod__(self, other: ValueCastType) -> 'Operator':
		return Operator('%', [self, other])

	def __rmod__(self, other: ValueCastType) -> 'Operator':
		return Operator('%', [other, self])

	def __floordiv__(self, other: ValueCastType) -> 'Operator':
		return Operator('//', [self, other])

	def __rfloordiv__(self, other: ValueCastType) -> 'Operator':
		return Operator('//', [other, self])


	def __check_shamt(self) -> None:
		if self.shape().signed:
			# Neither Python nor HDLs implement shifts by negative values; prohibit any shifts
			# by a signed value to make sure the shift amount can always be interpreted as
			# an unsigned value.
			raise TypeError('Shift amount must be unsigned')

	def __lshift__(self, other: ValueCastType) -> 'Operator':
		other = Value.cast(other)
		other.__check_shamt()
		return Operator('<<', [self, other])

	def __rlshift__(self, other: ValueCastType) -> 'Operator':
		self.__check_shamt()
		return Operator('<<', [other, self])

	def __rshift__(self, other: ValueCastType) -> 'Operator':
		other = Value.cast(other)
		other.__check_shamt()
		return Operator('>>', [self, other])

	def __rrshift__(self, other: ValueCastType) -> 'Operator':
		self.__check_shamt()
		return Operator('>>', [other, self])

	def __and__(self, other: ValueCastType) -> 'Operator':
		return Operator('&', [self, other])

	def __rand__(self, other: ValueCastType) -> 'Operator':
		return Operator('&', [other, self])

	def __xor__(self, other: ValueCastType) -> 'Operator':
		return Operator('^', [self, other])

	def __rxor__(self, other: ValueCastType) -> 'Operator':
		return Operator('^', [other, self])

	def __or__(self, other: ValueCastType) -> 'Operator':
		return Operator('|', [self, other])

	def __ror__(self, other: ValueCastType) -> 'Operator':
		return Operator('|', [other, self])

	def __eq__(self, other: ValueCastType) -> 'Operator':
		return Operator('==', [self, other])

	def __ne__(self, other: ValueCastType) -> 'Operator':
		return Operator('!=', [self, other])

	def __lt__(self, other: ValueCastType) -> 'Operator':
		return Operator('<', [self, other])

	def __le__(self, other: ValueCastType) -> 'Operator':
		return Operator('<=', [self, other])

	def __gt__(self, other: ValueCastType) -> 'Operator':
		return Operator('>', [self, other])

	def __ge__(self, other: ValueCastType) -> 'Operator':
		return Operator('>=', [self, other])

	def __abs__(self) -> 'Value':
		if self.shape().signed:
			return Mux(self >= 0, self, -self)[:len(self)]
		else:
			return self

	def __len__(self) -> int:
		return self.shape().width

	def __getitem__(self, key: Union[int, slice]) -> 'Slice':
		n = len(self)
		if isinstance(key, int):
			if key not in range(-n, n):
				raise IndexError(f'Index {key} is out of bounds for a {n}-bit value')
			if key < 0:
				key += n
			return Slice(self, key, key + 1, src_loc_at = 1)
		elif isinstance(key, slice):
			start, stop, step = key.indices(n)
			if step != 1:
				return Cat(self[i] for i in range(start, stop, step))
			return Slice(self, start, stop, src_loc_at = 1)
		else:
			raise TypeError(f'Cannot index value with {key!r}')

	def as_unsigned(self)  -> 'Operator':
		'''
		Conversion to unsigned.

		Returns
		-------
		Value, out
			This ``Value`` reinterpreted as a unsigned integer.

		'''

		return Operator('u', [self])

	def as_signed(self)  -> 'Operator':
		'''
		Conversion to signed.

		Returns
		-------
		Value, out
			This ``Value`` reinterpreted as a signed integer.

		'''

		return Operator('s', [self])

	def bool(self)  -> 'Operator':
		'''
		Conversion to boolean.

		Returns
		-------
		Value, out
			``1`` if any bits are set, ``0`` otherwise.

		'''

		return Operator('b', [self])

	def any(self)  -> 'Operator':
		'''
		Check if any bits are ``1``.

		Returns
		-------
		Value, out
			``1`` if any bits are set, ``0`` otherwise.

		'''

		return Operator('r|', [self])

	def all(self)  -> 'Operator':
		'''
		Check if all bits are ``1``.

		Returns
		-------
		Value, out
			``1`` if all bits are set, ``0`` otherwise.

		'''

		return Operator('r&', [self])

	def xor(self)  -> 'Operator':
		'''
		Compute pairwise exclusive-or of every bit.

		Returns
		-------
		Value, out
			``1`` if an odd number of bits are set, ``0`` if an even number of bits are set.

		'''

		return Operator('r^', [self])

	def implies(premise, conclusion: ValueCastType)  -> 'Operator':
		'''
		Implication.

		Returns
		-------
		Value, out
			``0`` if ``premise`` is true and ``conclusion`` is not, ``1`` otherwise.

		'''

		return ~premise | conclusion

	def bit_select(self, offset: Union['Value', int] , width: int) -> 'Part':
		'''
		Part-select with bit granularity.

		Selects a constant width but variable offset part of a ``Value``, such that successive
		parts overlap by all but 1 bit.

		Parameters
		----------
		offset : Value, int
			Index of first selected bit.
		width : int
			Number of selected bits.

		Returns
		-------
		Part, out
			Selected part of the ``Value``

		'''

		offset = Value.cast(offset)
		if type(offset) is Const and isinstance(width, int):
			return self[offset.value:offset.value + width]
		return Part(self, offset, width, stride = 1, src_loc_at = 1)

	def word_select(self, offset: Union['Value', int] , width: int) -> 'Part':
		'''
		Part-select with word granularity.

		Selects a constant width but variable offset part of a ``Value``, such that successive
		parts do not overlap.

		Parameters
		----------
		offset : Value, int
			Index of first selected word.
		width : int
			Number of selected bits.

		Returns
		-------
		Part, out
			Selected part of the ``Value``

		'''

		offset = Value.cast(offset)
		if type(offset) is Const and isinstance(width, int):
			return self[offset.value * width:(offset.value + 1) * width]
		return Part(self, offset, width, stride = width, src_loc_at = 1)

	def matches(self, *patterns: tuple[Union[int, str, Enum]]) -> 'Value':
		'''
		Pattern matching.

		Matches against a set of patterns, which may be integers or bit strings, recognizing
		the same grammar as ``Case()``.

		Parameters
		----------
		patterns : int, str, or Enum
			Patterns to match against.

		Returns
		-------
		Value, out
			``1`` if any pattern matches the value, ``0`` otherwise.

		'''

		matches = []
		for pattern in patterns:
			if not isinstance(pattern, (int, str, Enum)):
				raise SyntaxError(
					f'Match pattern must be an integer, a string, or an enumeration, not {pattern!r}'
				)

			if isinstance(pattern, str) and any(bit not in '01- \t' for bit in pattern):
				raise SyntaxError(
					f'Match pattern \'{pattern}\' must consist of 0, 1, and - (don\'t care) bits, and may include whitespace'
				)

			if (isinstance(pattern, str) and len(''.join(pattern.split())) != len(self)):
				raise SyntaxError(
					f'Match pattern \'{pattern}\' must have the same width as match value (which is {len(self)})'
				)

			if isinstance(pattern, int) and bits_for(pattern) > len(self):
				warnings.warn(
					f'Match pattern \'{pattern:b}\' is wider than match value (which has width {len(self)}); '
					'comparison will never be true',
					SyntaxWarning, stacklevel = 3
				)
				continue
			if isinstance(pattern, str):
				pattern = ''.join(pattern.split()) # remove whitespace
				mask    = int(pattern.replace('0', '1').replace('-', '0'), 2)
				pattern = int(pattern.replace('-', '0'), 2)
				matches.append((self & mask) == pattern)
			elif isinstance(pattern, int):
				matches.append(self == pattern)
			elif isinstance(pattern, Enum):
				matches.append(self == pattern.value)
			else:
				assert False
		if not matches:
			return Const(1)
		elif len(matches) == 1:
			return matches[0]
		else:
			return Cat(*matches).any()

	def shift_left(self, amount: int) -> 'Value':
		'''
		Shift left by constant amount.

		Parameters
		----------
		amount : int
			Amount to shift by.

		Returns
		-------
		Value, out
			If the amount is positive, the input shifted left. Otherwise, the input shifted right.

		'''

		if not isinstance(amount, int):
			raise TypeError(f'Shift amount must be an integer, not {amount!r}')
		if amount < 0:
			return self.shift_right(-amount)
		if self.shape().signed:
			return Cat(Const(0, amount), self).as_signed()
		else:
			return Cat(Const(0, amount), self) # unsigned

	def shift_right(self, amount: int) -> 'Value':
		'''
		Shift right by constant amount.

		Parameters
		----------
		amount : int
			Amount to shift by.

		Returns
		-------
		Value, out
			If the amount is positive, the input shifted right. Otherwise, the input shifted left.

		'''

		if not isinstance(amount, int):
			raise TypeError(f'Shift amount must be an integer, not {amount!r}')
		if amount < 0:
			return self.shift_left(-amount)
		if self.shape().signed:
			return self[amount:].as_signed()
		else:
			return self[amount:] # unsigned

	def rotate_left(self, amount: int) -> 'Value':
		'''
		Rotate left by constant amount.

		Parameters
		----------
		amount : int
			Amount to rotate by.

		Returns
		-------
		Value, out
			If the amount is positive, the input rotated left. Otherwise, the input rotated right.

		'''

		if not isinstance(amount, int):
			raise TypeError(f'Rotate amount must be an integer, not {amount!r}')
		if len(self) != 0:
			amount %= len(self)
		return Cat(self[-amount:], self[:-amount]) # meow :3

	def rotate_right(self, amount: int) -> 'Value':
		'''
		Rotate right by constant amount.

		Parameters
		----------
		amount : int
			Amount to rotate by.

		Returns
		-------
		Value, out
			If the amount is positive, the input rotated right. Otherwise, the input rotated right.

		'''

		if not isinstance(amount, int):
			raise TypeError(f'Rotate amount must be an integer, not {amount!r}')
		if len(self) != 0:
			amount %= len(self)
		return Cat(self[amount:], self[:amount])

	def replicate(self, count):
		'''
		Replication.

		A ``Value`` is replicated (repeated) several times to be used
		on the RHS of assignments::

			len(v.replicate(n)) == len(v) * n

		Parameters
		----------
		count : int
			Number of replications.

		Returns
		-------
		Value, out
			Replicated value.
		'''
		if not isinstance(count, int) or count < 0:
			raise TypeError(f'Replication count must be a non-negative integer, not {count!r}')
		return Cat(self for _ in range(count))

	def eq(self, value: 'Value') -> 'Assign':
		'''
		Assignment.

		Parameters
		----------
		value : Value, in
			Value to be assigned.

		Returns
		-------
		Assign
			Assignment statement that can be used in combinatorial or synchronous context.

		'''

		return Assign(self, value, src_loc_at = 1)

	@abstractmethod
	def shape(self) -> Shape:
		'''
		Bit width and signedness of a value.

		Returns
		-------
		Shape
			See :class:`Shape`.

		Examples
		--------
		>>> Signal(8).shape()
		Shape(width = 8, signed = False)
		>>> Const(0xaa).shape()
		Shape(width = 8, signed = False)

		'''

		raise NotImplementedError('.shape has not been implemented')

	def _lhs_signals(self):
		raise TypeError(f'Value {self!r} cannot be used in assignments')

	@abstractmethod
	def _rhs_signals(self):
		pass # :nocov:

	def _as_const(self):
		raise TypeError(f'Value {self!r} cannot be evaluated as constant')


@final
class Const(Value):
	'''
	A constant, literal integer value.

	Parameters
	----------
	value : int
	shape : int or tuple or None
		Either an integer ``width`` or a tuple ``(width, signed)`` specifying the number of bits
		in this constant and whether it is signed (can represent negative values).
		``shape`` defaults to the minimum possible width and signedness of ``value``.

	Attributes
	----------
	width : int
	signed : bool

	'''

	src_loc = None

	@staticmethod
	def normalize(value: int, shape: tuple[int, bool]):
		mask = (1 << shape.width) - 1
		value &= mask
		if shape.signed and (value >> (shape.width - 1)) & 1:
			value |= ~mask
		return value

	def __init__(
		self, value: int, shape: Optional[Union[int, tuple[int, bool]]] = None, *,
		src_loc_at: int = 0
	) -> None:
		# We deliberately do not call Value.__init__ here.
		self.value = int(value)
		if shape is None:
			shape = Shape(bits_for(self.value), signed = self.value < 0)
		elif isinstance(shape, int):
			shape = Shape(shape, signed = self.value < 0)
		else:
			if isinstance(shape, range) and self.value == shape.stop:
				warnings.warn(
					message =
						f'Value {self.value!r} equals the non-inclusive end of the constant '
						f'shape {shape!r}; this is likely an off-by-one error',
					category = SyntaxWarning,
					stacklevel = 2
				)
			shape = Shape.cast(shape, src_loc_at = 1 + src_loc_at)
		self.width  = shape.width
		self.signed = shape.signed
		self.value  = self.normalize(self.value, shape)

	def shape(self) -> Shape:
		return Shape(self.width, self.signed)

	def _rhs_signals(self) -> 'SignalSet':
		return SignalSet()

	def _as_const(self) -> int:
		return self.value

	def __repr__(self) -> str:
		return f'(const {self.width}\'{"s" if self.signed else ""}d{self.value})'


class AnyValue(Value, DUID):
	def __init__(
		self, shape: Union[Shape, int, tuple[int, bool], range, type, ShapeCastable] , *,
		src_loc_at: int = 0
	) -> None:
		super().__init__(src_loc_at = src_loc_at)
		shape = Shape.cast(shape, src_loc_at = 1 + src_loc_at)
		self.width  = shape.width
		self.signed = shape.signed

	def shape(self) -> Shape:
		return Shape(self.width, self.signed)

	def _rhs_signals(self) -> 'SignalSet':
		return SignalSet()


@final
class AnyConst(AnyValue):
	def __repr__(self) -> str:
		return f'(anyconst {self.width}\'{"s" if self.signed else ""})'


@final
class AnySeq(AnyValue):
	def __repr__(self) -> str:
		return f'(anyseq {self.width}\'{"s" if self.signed else ""})'


@final
class Operator(Value):
	def __init__(self, operator, operands , *, src_loc_at = 0) -> None:
		super().__init__(src_loc_at = 1 + src_loc_at)
		self.operator = operator
		self.operands = [Value.cast(op) for op in operands]

	def shape(self):
		def _bitwise_binary_shape(a_shape, b_shape):
			if not a_shape.signed and not b_shape.signed:
				# both operands unsigned
				return unsigned(max(a_shape.width, b_shape.width))
			elif a_shape.signed and b_shape.signed:
				# both operands signed
				return signed(max(a_shape.width, b_shape.width))
			elif not a_shape.signed and b_shape.signed:
				# first operand unsigned (add sign bit), second operand signed
				return signed(max(a_shape.width + 1, b_shape.width))
			else:
				# first signed, second operand unsigned (add sign bit)
				return signed(max(a_shape.width, b_shape.width + 1))

		op_shapes = list(map(lambda x: x.shape(), self.operands))
		if len(op_shapes) == 1:
			a_shape, = op_shapes
			if self.operator in ('+', '~'):
				return Shape(a_shape.width, a_shape.signed)
			if self.operator == '-':
				return Shape(a_shape.width + 1, True)
			if self.operator in ('b', 'r|', 'r&', 'r^'):
				return Shape(1, False)
			if self.operator == 'u':
				return Shape(a_shape.width, False)
			if self.operator == 's':
				return Shape(a_shape.width, True)
		elif len(op_shapes) == 2:
			a_shape, b_shape = op_shapes
			if self.operator == '+':
				o_shape = _bitwise_binary_shape(*op_shapes)
				return Shape(o_shape.width + 1, o_shape.signed)
			if self.operator == '-':
				o_shape = _bitwise_binary_shape(*op_shapes)
				return Shape(o_shape.width + 1, True)
			if self.operator == '*':
				return Shape(a_shape.width + b_shape.width, a_shape.signed or b_shape.signed)
			if self.operator == '//':
				return Shape(a_shape.width + b_shape.signed, a_shape.signed or b_shape.signed)
			if self.operator == '%':
				return Shape(b_shape.width, b_shape.signed)
			if self.operator in ('<', '<=', '==', '!=', '>', '>='):
				return Shape(1, False)
			if self.operator in ('&', '^', '|'):
				return _bitwise_binary_shape(*op_shapes)
			if self.operator == '<<':
				assert not b_shape.signed
				return Shape(a_shape.width + 2 ** b_shape.width - 1, a_shape.signed)
			if self.operator == '>>':
				assert not b_shape.signed
				return Shape(a_shape.width, a_shape.signed)
		elif len(op_shapes) == 3:
			if self.operator == 'm':
				s_shape, a_shape, b_shape = op_shapes
				return _bitwise_binary_shape(a_shape, b_shape)
		raise NotImplementedError(f'Operator {self.operator}/{len(op_shapes)} not implemented') # :nocov:

	def _lhs_signals(self):
		if self.operator in ('u', 's'):
			return union(op._lhs_signals() for op in self.operands)
		return super()._lhs_signals()

	def _rhs_signals(self):
		return union(op._rhs_signals() for op in self.operands)

	def __repr__(self):
		return f'({self.operator} {" ".join(map(repr, self.operands))})'


def Mux(sel: Value, val1: Value, val0: Value) -> Operator:
	'''
	Choose between two values.

	Parameters
	----------
	sel : Value, in
		Selector.
	val1 : Value, in
	val0 : Value, in
		Input values.

	Returns
	-------
	Value, out
		Output ``Value``. If ``sel`` is asserted, the Mux returns ``val1``, else ``val0``.

	'''

	return Operator('m', [sel, val1, val0])


@final
class Slice(Value):
	def __init__(
		self, value: ValueCastType, start: int, stop: int, *, src_loc_at: int = 0
	) -> None:
		if not isinstance(start, int):
			raise TypeError(f'Slice start must be an integer, not {start!r}')
		if not isinstance(stop, int):
			raise TypeError(f'Slice stop must be an integer, not {stop!r}')

		value = Value.cast(value)
		n = len(value)
		if start not in range(-n, n + 1):
			raise IndexError(f'Cannot start slice {start} bits into {n}-bit value')
		if start < 0:
			start += n
		if stop not in range(-n, n + 1):
			raise IndexError(f'Cannot stop slice {stop} bits into {n}-bit value')
		if stop < 0:
			stop += n
		if start > stop:
			raise IndexError(f'Slice start {start} must be less than slice stop {stop}')

		super().__init__(src_loc_at = src_loc_at)
		self.value = value
		self.start = int(start)
		self.stop  = int(stop)

	def shape(self) -> Shape:
		return Shape(self.stop - self.start)

	def _lhs_signals(self):
		return self.value._lhs_signals()

	def _rhs_signals(self):
		return self.value._rhs_signals()

	def __repr__(self) -> str:
		return f'(slice {repr(self.value)} {self.start}:{self.stop})'


@final
class Part(Value):
	def __init__(
		self, value: ValueCastType, offset: ValueCastType, width: int, stride: int = 1, *,
		src_loc_at: int = 0
	) -> None:
		if not isinstance(width, int) or width < 0:
			raise TypeError(f'Part width must be a non-negative integer, not {width!r}')
		if not isinstance(stride, int) or stride <= 0:
			raise TypeError(f'Part stride must be a positive integer, not {stride!r}')

		value  = Value.cast(value)
		offset = Value.cast(offset)
		if offset.shape().signed:
			raise TypeError('Part offset must be unsigned')

		super().__init__(src_loc_at = src_loc_at)
		self.value  = value
		self.offset = offset
		self.width  = width
		self.stride = stride

	def shape(self) -> Shape:
		return Shape(self.width)

	def _lhs_signals(self):
		return self.value._lhs_signals()

	def _rhs_signals(self):
		return self.value._rhs_signals() | self.offset._rhs_signals()

	def __repr__(self) -> str:
		return f'(part {repr(self.value)} {repr(self.offset)} {self.width} {self.stride})'


@final
class Cat(Value):
	'''
	Concatenate values.

	Form a compound ``Value`` from several smaller ones by concatenation.
	The first argument occupies the lower bits of the result.
	The return value can be used on either side of an assignment, that
	is, the concatenated value can be used as an argument on the RHS or
	as a target on the LHS. If it is used on the LHS, it must solely
	consist of ``Signal`` s, slices of ``Signal`` s, and other concatenations
	meeting these properties. The bit length of the return value is the sum of
	the bit lengths of the arguments::

		len(Cat(args)) == sum(len(arg) for arg in args)

	Parameters
	----------
	*args : Values or iterables of Values, inout
		``Value`` s to be concatenated.

	Returns
	-------
	Value, inout
		Resulting ``Value`` obtained by concatenation.

	'''

	def __init__(self, *args: Iterable[Value], src_loc_at: int = 0) -> None:
		super().__init__(src_loc_at = src_loc_at)
		self.parts = []
		for index, arg in enumerate(flatten(args)):
			if isinstance(arg, Enum) and (not isinstance(type(arg), ShapeCastable) or
				not hasattr(arg, '_torii_shape_')
			):
				warnings.warn(
					f'Argument #{index + 1} of \'Cat()\' is an enumerated value {arg!r} without '
					'a defined shape used in a bit vector context; use \'Const\' to specify '
					'the shape.',
					SyntaxWarning, stacklevel = 2 + src_loc_at
				)
			if isinstance(arg, int) and not isinstance(arg, Enum) and arg not in [0, 1]:
				warnings.warn(
					f'Argument #{index + 1} of \'Cat()\' is a bare integer {arg} used in bit vector '
					f'context; consider specifying the width explicitly using \'Const({arg}, {bits_for(arg)})\' '
					'instead',
					SyntaxWarning, stacklevel = 2 + src_loc_at
				)
			self.parts.append(Value.cast(arg))

	def shape(self) -> Shape:
		return Shape(sum(len(part) for part in self.parts))

	def _lhs_signals(self):
		return union((part._lhs_signals() for part in self.parts), start = SignalSet())

	def _rhs_signals(self):
		return union((part._rhs_signals() for part in self.parts), start = SignalSet())

	def _as_const(self) -> int:
		value = 0
		for part in reversed(self.parts):
			value <<= len(part)
			value |= part._as_const()
		return value

	def __repr__(self) -> str:
		return f'(cat {" ".join(map(repr, self.parts))})'


@deprecated('instead of `Repl(value, count)`, use `value.replicate(count)`')
def Repl(value: ValueCastType, count: int):
	'''
	Replicate a value

	An input value is replicated (repeated) several times
	to be used on the RHS of assignments::

		len(Repl(s, n)) == len(s) * n

	Parameters
	----------
	value : Value, in
		Input value to be replicated.
	count : int
		Number of replications.

	Returns
	-------
	Value, out
		Replicated value.

	'''
	if isinstance(value, int) and value not in [0, 1]:
		warnings.warn(
			f'Value argument of Repl() is a bare integer {value} used in bit vector '
			f'context; consider specifying explicit width using Const({value}, {bits_for(value)}) instead',
			SyntaxWarning, stacklevel = 3
		)

	return Value.cast(value).replicate(count)


# @final
class Signal(Value, DUID):
	'''
	A varying integer value.

	Parameters
	----------
	shape : ``Shape``-castable object or None
		Specification for the number of bits in this ``Signal`` and its signedness (whether it
		can represent negative values). See ``Shape.cast`` for details.
		If not specified, ``shape`` defaults to 1-bit and non-signed.
	name : str
		Name hint for this signal. If ``None`` (default) the name is inferred from the variable
		name this ``Signal`` is assigned to.
	reset : int or integral Enum
		Reset (synchronous) or default (combinatorial) value.
		When this ``Signal`` is assigned to in synchronous context and the corresponding clock
		domain is reset, the ``Signal`` assumes the given value. When this ``Signal`` is unassigned
		in combinatorial context (due to conditional assignments not being taken), the ``Signal``
		assumes its ``reset`` value. Defaults to 0.
	reset_less : bool
		If ``True``, do not generate reset logic for this ``Signal`` in synchronous statements.
		The ``reset`` value is only used as a combinatorial default or as the initial value.
		Defaults to ``False``.
	attrs : dict
		Dictionary of synthesis attributes.
	decoder : function or Enum
		A function converting integer signal values to human-readable strings (e.g. FSM state
		names). If an ``Enum`` subclass is passed, it is concisely decoded using format string
		``"{0.name:}/{0.value:}"``, or a number if the signal value is not a member of
		the enumeration.

	Attributes
	----------
	width : int
	signed : bool
	name : str
	reset : int
	reset_less : bool
	attrs : dict
	decoder : function

	'''

	def __init__(
		self, shape = None, *, name = None, reset = 0, reset_less = False,
		attrs = None, decoder = None, src_loc_at = 0
	):
		super().__init__(src_loc_at = src_loc_at)

		if name is not None and not isinstance(name, str):
			raise TypeError(f'Name must be a string, not {name!r}')
		self.name = name or tracer.get_var_name(depth = 2 + src_loc_at, default = '$signal')

		orig_shape = shape
		if shape is None:
			shape = unsigned(1)
		else:
			shape = Shape.cast(shape, src_loc_at = 1 + src_loc_at)
		self.width  = shape.width
		self.signed = shape.signed

		orig_reset = reset
		try:
			reset = Const.cast(reset)
		except TypeError:
			raise TypeError(
				f'Reset value must be a constant-castable expression, not {orig_reset!r}'
			)

		# Avoid false positives for all-zeroes and all-ones
		if orig_reset not in (0, -1):
			if reset.shape().signed and not self.signed:
				warnings.warn(
					message = f'Reset value {orig_reset!r} is signed, but the signal shape is {shape!r}',
					category = SyntaxWarning,
					stacklevel = 2
				)
			elif (
				reset.shape().width > self.width or
				reset.shape().width == self.width and
				self.signed and not reset.shape().signed
			):
				warnings.warn(
					message = f'Reset value {orig_reset!r} will be truncated to the signal shape {shape!r}',
					category = SyntaxWarning,
					stacklevel = 2
				)
		self.reset = reset.value
		self.reset_less = bool(reset_less)

		if isinstance(orig_shape, range) and self.reset == orig_shape.stop:
			warnings.warn(
				message =
					f'Reset value {self.reset!r} equals the non-inclusive end of the signal '
					f'shape {orig_shape!r}; this is likely an off-by-one error',
				category = SyntaxWarning,
				stacklevel = 2
			)

		self.attrs = OrderedDict(() if attrs is None else attrs)

		if decoder is None and isinstance(orig_shape, type) and issubclass(orig_shape, Enum):
			decoder = orig_shape
		if isinstance(decoder, type) and issubclass(decoder, Enum):
			def enum_decoder(value):
				try:
					return '{0.name:}/{0.value:}'.format(decoder(value))
				except ValueError:
					return str(value)
			self.decoder = enum_decoder
			self._enum_class = decoder
		else:
			self.decoder = decoder
			self._enum_class = None

	# Not a @classmethod because torii.compat requires it.
	@staticmethod
	def like(other, *, name = None, name_suffix = None, src_loc_at = 0, **kwargs):
		'''
		Create Signal based on another.

		Parameters
		----------
		other : Value
			Object to base this Signal on.

		'''

		if name is not None:
			new_name = str(name)
		elif name_suffix is not None:
			new_name = other.name + str(name_suffix)
		else:
			new_name = tracer.get_var_name(depth = 2 + src_loc_at, default = '$like')

		if isinstance(other, ValueCastable):
			shape = other.shape()
		else:
			shape = Value.cast(other).shape()
		kw = dict(shape = shape, name = new_name)
		if isinstance(other, Signal):
			kw.update(
				reset = other.reset, reset_less = other.reset_less,
				attrs = other.attrs, decoder = other.decoder
			)
		kw.update(kwargs)
		return Signal(**kw, src_loc_at = 1 + src_loc_at)

	def shape(self):
		return Shape(self.width, self.signed)

	def _lhs_signals(self):
		return SignalSet((self,))

	def _rhs_signals(self):
		return SignalSet((self,))

	def __repr__(self):
		return f'(sig {self.name})'


@final
class ClockSignal(Value):
	'''
	Clock signal for a clock domain.

	Any ``ClockSignal`` is equivalent to ``cd.clk`` for a clock domain with the corresponding name.
	All of these signals ultimately refer to the same signal, but they can be manipulated
	independently of the clock domain, even before the clock domain is created.

	Parameters
	----------
	domain : str
		Clock domain to obtain a clock signal for. Defaults to ``'sync'``.

	'''

	def __init__(self, domain = 'sync', *, src_loc_at = 0):
		super().__init__(src_loc_at = src_loc_at)
		if not isinstance(domain, str):
			raise TypeError(f'Clock domain name must be a string, not {domain!r}')
		if domain == 'comb':
			raise ValueError(f'Domain \'{domain}\' does not have a clock')
		self.domain = domain

	def shape(self):
		return Shape(1)

	def _lhs_signals(self):
		return SignalSet((self,))

	def _rhs_signals(self):
		raise NotImplementedError('ClockSignal must be lowered to a concrete signal') # :nocov:

	def __repr__(self):
		return f'(clk {self.domain})'


@final
class ResetSignal(Value):
	'''
	Reset signal for a clock domain.

	Any ``ResetSignal`` is equivalent to ``cd.rst`` for a clock domain with the corresponding name.
	All of these signals ultimately refer to the same signal, but they can be manipulated
	independently of the clock domain, even before the clock domain is created.

	Parameters
	----------
	domain : str
		Clock domain to obtain a reset signal for. Defaults to ``'sync'``.
	allow_reset_less : bool
		If the clock domain is reset-less, act as a constant ``0`` instead of reporting an error.

	'''

	def __init__(self, domain: str = 'sync', allow_reset_less: bool = False, *, src_loc_at: int = 0):
		super().__init__(src_loc_at = src_loc_at)
		if not isinstance(domain, str):
			raise TypeError(f'Clock domain name must be a string, not {domain!r}')
		if domain == 'comb':
			raise ValueError(f'Domain \'{domain}\' does not have a reset')
		self.domain = domain
		self.allow_reset_less = allow_reset_less

	def shape(self) -> Shape:
		return Shape(1)

	def _lhs_signals(self) -> 'SignalSet':
		return SignalSet((self,))

	def _rhs_signals(self) -> None:
		raise NotImplementedError('ResetSignal must be lowered to a concrete signal') # :nocov:

	def __repr__(self) -> str:
		return f'(rst {self.domain})'


class Array(MutableSequence):
	'''
	Addressable multiplexer.

	An array is similar to a ``list`` that can also be indexed by ``Value``s;
	indexing by an integer or a slice works the same as for Python lists,
	but indexing by a ``Value`` results in a proxy.

	The array proxy can be used as an ordinary ``Value``, i.e. participate in calculations and
	assignments, provided that all elements of the array are values. The array proxy also supports
	attribute access and further indexing, each returning another array proxy; this means that
	the results of indexing into arrays, arrays of records, and arrays of arrays can all
	be used as first-class values.

	It is an error to change an array or any of its elements after an array proxy was created.
	Changing the array directly will raise an exception. However, it is not possible to detect
	the elements being modified; if an element's attribute or element is modified after the proxy
	for it has been created, the proxy will refer to stale data.

	Examples
	--------

	Simple array::

		gpios = Array(Signal() for _ in range(10))
		with m.If(bus.we):
			m.d.sync += gpios[bus.addr].eq(bus.w_data)
		with m.Else():
			m.d.sync += bus.r_data.eq(gpios[bus.addr])

	Multidimensional array::

		mult = Array(Array(x * y for y in range(10)) for x in range(10))
		a = Signal.range(10)
		b = Signal.range(10)
		r = Signal(8)
		m.d.comb += r.eq(mult[a][b])

	Array of records::

		layout = [
			("r_data", 16),
			("r_en",   1),
		]
		buses  = Array(Record(layout) for busno in range(4))
		master = Record(layout)
		m.d.comb += [
			buses[sel].r_en.eq(master.r_en),
			master.r_data.eq(buses[sel].r_data),
		]

	'''

	def __init__(self, iterable = ()):
		self._inner    = list(iterable)
		self._proxy_at = None
		self._mutable  = True

	def __getitem__(self, index):
		if isinstance(index, Value):
			if self._mutable:
				self._proxy_at = tracer.get_src_loc()
				self._mutable  = False
			return ArrayProxy(self, index)
		else:
			return self._inner[index]

	def __len__(self):
		return len(self._inner)

	def _check_mutability(self):
		if not self._mutable:
			raise ValueError('Array can no longer be mutated after it was indexed with a value at {}:{}'.format(*self._proxy_at))

	def __setitem__(self, index, value):
		self._check_mutability()
		self._inner[index] = value

	def __delitem__(self, index):
		self._check_mutability()
		del self._inner[index]

	def insert(self, index, value):
		self._check_mutability()
		self._inner.insert(index, value)

	def __repr__(self):
		return f'(array{" mutable" if self._mutable else ""} [{", ".join(map(repr, self._inner))}])'

@final
class ArrayProxy(Value):
	def __init__(self, elems, index, *, src_loc_at = 0):
		super().__init__(src_loc_at = 1 + src_loc_at)
		self.elems = elems
		self.index = Value.cast(index)

	def __getattr__(self, attr):
		return ArrayProxy([getattr(elem, attr) for elem in self.elems], self.index)

	def __getitem__(self, index):
		return ArrayProxy([        elem[index] for elem in self.elems], self.index)

	def _iter_as_values(self):
		return (Value.cast(elem) for elem in self.elems)

	def shape(self):
		unsigned_width = signed_width = 0
		has_unsigned = has_signed = False
		for elem_shape in (elem.shape() for elem in self._iter_as_values()):
			if elem_shape.signed:
				has_signed = True
				signed_width = max(signed_width, elem_shape.width)
			else:
				has_unsigned = True
				unsigned_width = max(unsigned_width, elem_shape.width)
		# The shape of the proxy must be such that it preserves the mathematical value of the array
		# elements. I.e., shape-wise, an array proxy must be identical to an equivalent mux tree.
		# To ensure this holds, if the array contains both signed and unsigned values, make sure
		# that every unsigned value is zero-extended by at least one bit.
		if has_signed and has_unsigned and unsigned_width >= signed_width:
			# Array contains both signed and unsigned values, and at least one of the unsigned
			# values won't be zero-extended otherwise.
			return signed(unsigned_width + 1)
		else:
			# Array contains values of the same signedness, or else all of the unsigned values
			# are zero-extended.
			return Shape(max(unsigned_width, signed_width), has_signed)

	def _lhs_signals(self):
		signals = union((elem._lhs_signals() for elem in self._iter_as_values()),
						start = SignalSet())
		return signals

	def _rhs_signals(self):
		signals = union((elem._rhs_signals() for elem in self._iter_as_values()),
						start = SignalSet())
		return self.index._rhs_signals() | signals

	def __repr__(self):
		return f'(proxy (array [{", ".join(map(repr, self.elems))}]) {self.index!r})'

class ValueCastable:
	'''
	Interface of user-defined objects that can be cast to :class:`Value` s.

	An object deriving from :class:`ValueCastable`` is automatically converted to a :class:`Value`
	when it is used in a context where a :class:`Value`` is expected. Such objects can implement
	different or richer semantics than what is supported by the core Torii language, yet still
	be transparently used with it as long as the final underlying representation is a single
	Torii :class:`Value`. These objects also need not commit to a specific representation until
	they are converted to a concrete Torii value.

	Note that it is necessary to ensure that Torii's view of representation of all values stays
	internally consistent. The class deriving from :class:`ValueCastable`` must decorate
	the :meth:`as_value` method with the :meth:`lowermethod` decorator, which ensures that all
	calls to :meth:`as_value` return the same :class:`Value` representation. If the class deriving
	from :class:`ValueCastable` is mutable, it is up to the user to ensure that it is not mutated
	in a way that changes its representation after the first call to :meth:`as_value`.

	'''

	def __init_subclass__(cls, **kwargs):
		if not hasattr(cls, 'as_value'):
			raise TypeError(
				f'Class \'{cls.__name__}\' deriving from `ValueCastable` must override the `as_value` method'
			)
		if not hasattr(cls, 'shape'):
			raise TypeError(
				f'Class \'{cls.__name__}\' deriving from `ValueCastable` must override the `shape` method'
			)
		if not hasattr(cls.as_value, '_ValueCastable__memoized'):
			raise TypeError(
				f'Class \'{cls.__name__}\' deriving from `ValueCastable` must decorate '
				'the `as_value` method with the `ValueCastable.lowermethod` decorator'
			)

	@staticmethod
	def lowermethod(func):
		'''
		Decorator to memoize lowering methods.

		Ensures the decorated method is called only once, with subsequent method calls returning
		the object returned by the first first method call.

		This decorator is required to decorate the ``as_value`` method of ``ValueCastable``
		subclasses. This is to ensure that Torii's view of representation of all values stays
		internally consistent.

		'''

		@functools.wraps(func)
		def wrapper_memoized(self, *args, **kwargs):
			# Use `in self.__dict__` instead of `hasattr` to avoid interfering with custom
			# `__getattr__` implementations.
			if '_ValueCastable__lowered_to' not in self.__dict__:
				self.__lowered_to = func(self, *args, **kwargs)
			return self.__lowered_to
		wrapper_memoized.__memoized = True
		return wrapper_memoized


@final
class Sample(Value):
	'''
	Value from the past.

	A ``Sample`` of an expression is equal to the value of the expression ``clocks`` clock edges
	of the ``domain`` clock back. If that moment is before the beginning of time, it is equal
	to the value of the expression calculated as if each signal had its reset value.

	'''

	def __init__(self, expr, clocks, domain, *, src_loc_at = 0):
		super().__init__(src_loc_at = 1 + src_loc_at)
		self.value  = Value.cast(expr)
		self.clocks = int(clocks)
		self.domain = domain
		if not isinstance(self.value, (Const, Signal, ClockSignal, ResetSignal, Initial)):
			raise TypeError(f'Sampled value must be a signal or a constant, not {self.value!r}')
		if self.clocks < 0:
			raise ValueError(f'Cannot sample a value {-self.clocks} cycles in the future')
		if not (self.domain is None or isinstance(self.domain, str)):
			raise TypeError(f'Domain name must be a string or None, not {self.domain!r}')

	def shape(self):
		return self.value.shape()

	def _rhs_signals(self):
		return SignalSet((self,))

	def __repr__(self):
		return f'(sample {self.value!r} @ {"<default>" if self.domain is None else self.domain}[{self.clocks}])'


def Past(expr, clocks = 1, domain = None):
	return Sample(expr, clocks, domain)


def Stable(expr, clocks = 0, domain = None):
	return Sample(expr, clocks + 1, domain) == Sample(expr, clocks, domain)


def Rose(expr, clocks = 0, domain = None):
	return ~Sample(expr, clocks + 1, domain) & Sample(expr, clocks, domain)


def Fell(expr, clocks = 0, domain = None):
	return Sample(expr, clocks + 1, domain) & ~Sample(expr, clocks, domain)


@final
class Initial(Value):
	'''
	Start indicator, for model checking.

	An ``Initial`` signal is ``1`` at the first cycle of model checking, and ``0`` at any other.

	'''

	def __init__(self, *, src_loc_at = 0):
		super().__init__(src_loc_at = src_loc_at)

	def shape(self):
		return Shape(1)

	def _rhs_signals(self):
		return SignalSet((self,))

	def __repr__(self):
		return '(initial)'


class _StatementList(list):
	def __repr__(self):
		return f'({" ".join(map(repr, self))})'


class Statement:
	def __init__(self, *, src_loc_at = 0):
		self.src_loc = tracer.get_src_loc(1 + src_loc_at)

	@staticmethod
	def cast(obj):
		if isinstance(obj, Iterable):
			return _StatementList(list(chain.from_iterable(map(Statement.cast, obj))))
		else:
			if isinstance(obj, Statement):
				return _StatementList([obj])
			else:
				raise TypeError(f'Object {obj!r} is not an Torii statement')


@final
class Assign(Statement):
	def __init__(self, lhs, rhs, *, src_loc_at = 0):
		super().__init__(src_loc_at = src_loc_at)
		self.lhs = Value.cast(lhs)
		self.rhs = Value.cast(rhs)

	def _lhs_signals(self):
		return self.lhs._lhs_signals()

	def _rhs_signals(self):
		return self.lhs._rhs_signals() | self.rhs._rhs_signals()

	def __repr__(self):
		return f'(eq {self.lhs!r} {self.rhs!r})'


class UnusedProperty(UnusedMustUse):
	pass


class Property(Statement, MustUse):
	_MustUse__warning = UnusedProperty

	def __init__(
		self, test: ValueCastType, *, _check: Optional[Signal] = None, _en: Optional[Signal] = None,
		name: Optional[str] = None, src_loc_at: int = 0
	) -> None:
		super().__init__(src_loc_at = src_loc_at)
		self.test   = Value.cast(test)
		self._check = _check
		self._en    = _en
		self.name   = name
		if not isinstance(self.name, str) and self.name is not None:
			raise TypeError(f'Property name must be a string of None, not {self.name!r}')

		if self._check is None:
			self._check = Signal(reset_less = True, name = f'${self._kind}$check')
			self._check.src_loc = self.src_loc
		if _en is None:
			self._en = Signal(reset_less = True, name = f'${self._kind}$en')
			self._en.src_loc = self.src_loc

	def _lhs_signals(self):
		return SignalSet((self._en, self._check))

	def _rhs_signals(self):
		return self.test._rhs_signals()

	def __repr__(self) -> str:
		if self.name is not None:
			return f'({self.name}: {self._kind} {self.test!r})'
		return f'({self._kind} {self.test!r})'


@final
class Assert(Property):
	_kind = 'assert'


@final
class Assume(Property):
	_kind = 'assume'


@final
class Cover(Property):
	_kind = 'cover'


# @final
class Switch(Statement):
	def __init__(self, test, cases, *, src_loc = None, src_loc_at = 0, case_src_locs = {}):
		if src_loc is None:
			super().__init__(src_loc_at = src_loc_at)
		else:
			# Switch is a bit special in terms of location tracking because it is usually created
			# long after the control has left the statement that directly caused its creation.
			self.src_loc = src_loc
		# Switch is also a bit special in that its parts also have location information. It can't
		# be automatically traced, so whatever constructs a Switch may optionally provide it.
		self.case_src_locs = {}

		self.test  = Value.cast(test)
		self.cases = OrderedDict()
		for orig_keys, stmts in cases.items():
			# Map: None -> (); key -> (key,); (key...) -> (key...)
			keys = orig_keys
			if keys is None:
				keys = ()
			if not isinstance(keys, tuple):
				keys = (keys,)
			# Map: 2 -> "0010"; "0010" -> "0010"
			new_keys = ()
			key_mask = (1 << len(self.test)) - 1
			for key in keys:
				if isinstance(key, str):
					key = "".join(key.split()) # remove whitespace
				elif isinstance(key, int):
					key = format(key & key_mask, 'b').rjust(len(self.test), '0')
				elif isinstance(key, Enum):
					key = format(key.value & key_mask, 'b').rjust(len(self.test), '0')
				else:
					raise TypeError(f'Object {key!r} cannot be used as a switch key')
				assert len(key) == len(self.test)
				new_keys = (*new_keys, key)
			if not isinstance(stmts, Iterable):
				stmts = [stmts]
			self.cases[new_keys] = Statement.cast(stmts)
			if orig_keys in case_src_locs:
				self.case_src_locs[new_keys] = case_src_locs[orig_keys]

	def _lhs_signals(self):
		signals = union((s._lhs_signals() for ss in self.cases.values() for s in ss),
						start = SignalSet())
		return signals

	def _rhs_signals(self):
		signals = union((s._rhs_signals() for ss in self.cases.values() for s in ss),
						start = SignalSet())
		return self.test._rhs_signals() | signals

	def __repr__(self) -> str:
		def case_repr(keys, stmts):
			stmts_repr = ' '.join(map(repr, stmts))
			if keys == ():
				return f'(default {stmts_repr})'
			elif len(keys) == 1:
				return f'(case {keys[0]} {stmts_repr})'
			else:
				return f'(case ({" ".join(keys)}) {stmts_repr})'
		case_reprs = [ case_repr(keys, stmts) for keys, stmts in self.cases.items() ]
		return f'(switch {self.test!r} {" ".join(case_reprs)})'


class _MappedKeyCollection(metaclass = ABCMeta):
	@abstractmethod
	def _map_key(self, key):
		pass # :nocov:

	@abstractmethod
	def _unmap_key(self, key):
		pass # :nocov:


class _MappedKeyDict(MutableMapping, _MappedKeyCollection):
	def __init__(self, pairs = ()):
		self._storage = OrderedDict()
		for key, value in pairs:
			self[key] = value

	def __getitem__(self, key):
		key = None if key is None else self._map_key(key)
		return self._storage[key]

	def __setitem__(self, key, value):
		key = None if key is None else self._map_key(key)
		self._storage[key] = value

	def __delitem__(self, key):
		key = None if key is None else self._map_key(key)
		del self._storage[key]

	def __iter__(self):
		for key in self._storage:
			if key is None:
				yield None
			else:
				yield self._unmap_key(key)

	def __eq__(self, other):
		if not isinstance(other, type(self)):
			return False
		if len(self) != len(other):
			return False
		for ak, bk in zip(sorted(self._storage), sorted(other._storage)):
			if ak != bk:
				return False
			if self._storage[ak] != other._storage[bk]:
				return False
		return True

	def __len__(self):
		return len(self._storage)

	def __repr__(self):
		pairs = [ f'({k!r}, {v!r})' for k, v in self.items() ]
		return f'{type(self).__module__}.{type(self).__name__}([{", ".join(pairs)}])'


class _MappedKeySet(MutableSet, _MappedKeyCollection):
	def __init__(self, elements = ()) -> None:
		self._storage = OrderedDict()
		for elem in elements:
			self.add(elem)

	def add(self, value) -> None:
		self._storage[self._map_key(value)] = None

	def update(self, values) -> None:
		for value in values:
			self.add(value)

	def discard(self, value) -> None:
		if value in self:
			del self._storage[self._map_key(value)]

	def __contains__(self, value) -> bool:
		return self._map_key(value) in self._storage

	def __iter__(self):
		for key in [ k for k in self._storage ]:
			yield self._unmap_key(key)

	def __len__(self) -> int:
		return len(self._storage)

	def __repr__(self) -> str:
		return f'{type(self).__module__}.{type(self).__name__}({", ".join(repr(x) for x in self)})'


class ValueKey:
	def __init__(self, value: ValueCastType) -> None:
		self.value = Value.cast(value)
		if isinstance(self.value, Const):
			self._hash = hash(self.value.value)
		elif isinstance(self.value, (Signal, AnyValue)):
			self._hash = hash(self.value.duid)
		elif isinstance(self.value, (ClockSignal, ResetSignal)):
			self._hash = hash(self.value.domain)
		elif isinstance(self.value, Operator):
			self._hash = hash((
				self.value.operator,
				tuple(ValueKey(o) for o in self.value.operands)
			))
		elif isinstance(self.value, Slice):
			self._hash = hash((ValueKey(self.value.value), self.value.start, self.value.stop))
		elif isinstance(self.value, Part):
			self._hash = hash((
				ValueKey(self.value.value), ValueKey(self.value.offset),
				self.value.width, self.value.stride
			))
		elif isinstance(self.value, Cat):
			self._hash = hash(tuple(ValueKey(o) for o in self.value.parts))
		elif isinstance(self.value, ArrayProxy):
			self._hash = hash((
				ValueKey(self.value.index),
				tuple(ValueKey(e) for e in self.value._iter_as_values())
			))
		elif isinstance(self.value, Sample):
			self._hash = hash((ValueKey(self.value.value), self.value.clocks, self.value.domain))
		elif isinstance(self.value, Initial):
			self._hash = 0
		else: # :nocov:
			raise TypeError(f'Object {self.value!r} cannot be used as a key in value collections')

	def __hash__(self) -> int:
		return self._hash

	def __eq__(self, other: 'ValueKey'):
		if not isinstance(other, ValueKey):
			return False
		if not isinstance(self.value, type(other.value)):
			return False

		if isinstance(self.value, Const):
			return (
				self.value.value == other.value.value and
				self.value.width == other.value.width
			)
		elif isinstance(self.value, (Signal, AnyValue)):
			return self.value is other.value
		elif isinstance(self.value, (ClockSignal, ResetSignal)):
			return self.value.domain == other.value.domain
		elif isinstance(self.value, Operator):
			return (
				self.value.operator == other.value.operator and
				len(self.value.operands) == len(other.value.operands) and
				all(
					ValueKey(a) == ValueKey(b)
					for a, b in zip(self.value.operands, other.value.operands)
				)
			)
		elif isinstance(self.value, Slice):
			return (
				ValueKey(self.value.value) == ValueKey(other.value.value) and
				self.value.start == other.value.start and
				self.value.stop == other.value.stop
			)
		elif isinstance(self.value, Part):
			return (
				ValueKey(self.value.value) == ValueKey(other.value.value) and
				ValueKey(self.value.offset) == ValueKey(other.value.offset) and
				self.value.width == other.value.width and
				self.value.stride == other.value.stride
			)
		elif isinstance(self.value, Cat):
			return (
				len(self.value.parts) == len(other.value.parts) and
				all(
					ValueKey(a) == ValueKey(b)
					for a, b in zip(self.value.parts, other.value.parts)
				)
			)
		elif isinstance(self.value, ArrayProxy):
			return (
				ValueKey(self.value.index) == ValueKey(other.value.index) and
				len(self.value.elems) == len(other.value.elems) and
				all(
					ValueKey(a) == ValueKey(b)
					for a, b in zip(self.value._iter_as_values(), other.value._iter_as_values())
				)
			)
		elif isinstance(self.value, Sample):
			return (
				ValueKey(self.value.value) == ValueKey(other.value.value) and
				self.value.clocks == other.value.clocks and
				self.value.domain == self.value.domain
			)
		elif isinstance(self.value, Initial):
			return True
		else: # :nocov:
			raise TypeError(f'Object {self.value!r} cannot be used as a key in value collections')

	def __lt__(self, other: 'ValueKey') -> bool:
		if not isinstance(other, ValueKey):
			return False
		if not isinstance(self.value, type(other.value)):
			return False

		if isinstance(self.value, Const):
			return self.value < other.value
		elif isinstance(self.value, (Signal, AnyValue)):
			return self.value.duid < other.value.duid
		elif isinstance(self.value, Slice):
			return (
				ValueKey(self.value.value) < ValueKey(other.value.value) and
				self.value.start < other.value.start and
				self.value.end < other.value.end
			)
		else: # :nocov:
			raise TypeError(f'Object {other!r} cannot be used as a key in value collections')

	def __repr__(self) -> str:
		return f'<{__name__}.ValueKey {self.value!r}>'


class ValueDict(_MappedKeyDict):
	_map_key = ValueKey

	def _unmap_key(self, key):
		return key.value


class ValueSet(_MappedKeySet):
	_map_key = ValueKey

	def _unmap_key(self, key):
		return key.value


class SignalKey:
	def __init__(self, signal: Union[Signal, ClockSignal, ResetSignal]) -> None:
		self.signal = signal
		if isinstance(signal, Signal):
			self._intern = (0, signal.duid)
		elif type(signal) is ClockSignal:
			self._intern = (1, signal.domain)
		elif type(signal) is ResetSignal:
			self._intern = (2, signal.domain)
		else:
			raise TypeError(f'Object {signal!r} is not an Torii signal')

	def __hash__(self) -> int:
		return hash(self._intern)

	def __eq__(self, other: 'SignalKey') -> bool:
		if type(other) is not SignalKey:
			return False
		return self._intern == other._intern

	def __lt__(self, other: 'SignalKey') -> bool:
		if type(other) is not SignalKey:
			raise TypeError(f'Object {other!r} cannot be compared to a SignalKey')
		return self._intern < other._intern

	def __repr__(self) -> str:
		return f'<{__name__}.SignalKey {self.signal!r}>'


class SignalDict(_MappedKeyDict):
	_map_key = SignalKey

	def _unmap_key(self, key):
		return key.signal


class SignalSet(_MappedKeySet):
	_map_key = SignalKey

	def _unmap_key(self, key):
		return key.signal



# SPDX-License-Identifier: BSD-2-Clause

from abc        import abstractmethod
from typing     import Literal, Optional

from ....build  import Attrs, Clock, Subsignal, TemplatedPlatform
from ....hdl    import (
	ClockDomain, ClockSignal, Const, Instance, Module, Record, Signal
)
from ....lib.io import Pin

__all__ = (
	'MachXO2Or3LPlatform',
	'MachXO2Platform',
	'MachXO3LPlatform',
)


# MachXO2 and MachXO3L primitives are the same. Handle both using
# one class and expose user-aliases for convenience.
class MachXO2Or3LPlatform(TemplatedPlatform):
	'''
	Required tools:
		* ``pnmainc``
		* ``ddtcmd``

	The environment is populated by running the script specified in the environment variable
	``TORII_ENV_DIAMOND``, if present. On Linux, diamond_env as provided by Diamond
	itself is a good candidate. On Windows, the following script (named ``diamond_env.bat``,
	for instance) is known to work::

		@echo off
		set PATH=C:\\lscc\\diamond\\%DIAMOND_VERSION%\\bin\\nt64;%PATH%

	Available overrides:
		* ``script_project``: inserts commands before ``prj_project save`` in Tcl script.
		* ``script_after_export``: inserts commands after ``prj_run Export`` in Tcl script.
		* ``add_preferences``: inserts commands at the end of the LPF file.
		* ``add_constraints``: inserts commands at the end of the XDC file.

	Build products:
		* ``{{name}}_impl/{{name}}_impl.htm``: consolidated log.
		* ``{{name}}.jed``: JEDEC fuse file.
		* ``{{name}}.bit``: binary bitstream.
		* ``{{name}}.svf``: JTAG programming vector for FLASH programming.
		* ``{{name}}_flash.svf``: JTAG programming vector for FLASH programming.
		* ``{{name}}_sram.svf``: JTAG programming vector for SRAM programming.

	'''

	toolchain = 'Diamond'

	@property
	@abstractmethod
	def device(self) -> str:
		raise NotImplementedError('Platform must implement this property')

	@property
	@abstractmethod
	def package(self) -> str:
		raise NotImplementedError('Platform must implement this property')

	@property
	@abstractmethod
	def speed(self) -> str:
		raise NotImplementedError('Platform must implement this property')

	grade   = 'C' # [C]ommercial, [I]ndustrial

	required_tools = [
		'pnmainc',
		'ddtcmd'
	]
	file_templates = {
		**TemplatedPlatform.build_script_templates,
		'build_{{name}}.sh': r'''
			# {{autogenerated}}
			set -e{{verbose("x")}}
			if [ -z "$BASH" ] ; then exec /bin/bash "$0" "$@"; fi
			if [ -n "${{platform._toolchain_env_var}}" ]; then
				bindir=$(dirname "${{platform._toolchain_env_var}}")
				. "${{platform._toolchain_env_var}}"
			fi
			{{emit_commands("sh")}}
		''',
		'{{name}}.v': r'''
			/* {{autogenerated}} */
			{{emit_verilog()}}
		''',
		'{{name}}.debug.v': r'''
			/* {{autogenerated}} */
			{{emit_debug_verilog()}}
		''',
		'{{name}}.tcl': r'''
			prj_project new -name {{name}} -impl impl -impl_dir {{name}}_impl \
				-dev {{platform.device}}-{{platform.speed}}{{platform.package}}{{platform.grade}} \
				-lpf {{name}}.lpf \
				-synthesis synplify
			{% for file in platform.iter_files(".v", ".sv", ".vhd", ".vhdl") -%}
				prj_src add {{file|tcl_escape}}
			{% endfor %}
			prj_src add {{name}}.v
			prj_impl option top {{name}}
			prj_src add {{name}}.sdc
			{{get_override("script_project")|default("# (script_project placeholder)")}}
			prj_project save
			prj_run Synthesis -impl impl
			prj_run Translate -impl impl
			prj_run Map -impl impl
			prj_run PAR -impl impl
			prj_run Export -impl impl -task Bitgen
			prj_run Export -impl impl -task Jedecgen
			{{get_override("script_after_export")|default("# (script_after_export placeholder)")}}
		''',
		'{{name}}.lpf': r'''
			# {{autogenerated}}
			BLOCK ASYNCPATHS;
			BLOCK RESETPATHS;
			{% for port_name, pin_name, attrs in platform.iter_port_constraints_bits() -%}
				LOCATE COMP '{{port_name}}" SITE '{{pin_name}}";
				{% if attrs -%}
				IOBUF PORT '{{port_name}}"
					{%- for key, value in attrs.items() %} {{key}}={{value}}{% endfor %};
				{% endif %}
			{% endfor %}
			{{get_override("add_preferences")|default("# (add_preferences placeholder)")}}
		''',
		'{{name}}.sdc': r'''
			{% for net_signal, port_signal, frequency in platform.iter_clock_constraints() -%}
				{% if port_signal is not none -%}
					create_clock -name {{port_signal.name|tcl_escape}} -period {{1000000000/frequency}} [get_ports {{port_signal.name|tcl_escape}}]
				{% else -%}
					create_clock -name {{net_signal.name|tcl_escape}} -period {{1000000000/frequency}} [get_nets {{net_signal|hierarchy("/")|tcl_escape}}]
				{% endif %}
			{% endfor %}
			{{get_override("add_constraints")|default("# (add_constraints placeholder)")}}
		''' # noqa: E501
	}
	command_templates = [
		# These don't have any usable command-line option overrides.
		r'''
		{{invoke_tool("pnmainc")}}
			{{name}}.tcl
		''',
		r'''
		{{invoke_tool("ddtcmd")}}
			-oft -bit
			-if {{name}}_impl/{{name}}_impl.bit -of {{name}}.bit
		''',
		r'''
		{{invoke_tool("ddtcmd")}}
			-oft -jed
			-dev {{platform.device}}-{{platform.speed}}{{platform.package}}{{platform.grade}}
			-if {{name}}_impl/{{name}}_impl.jed -of {{name}}.jed
		''',
		r'''
		{{invoke_tool("ddtcmd")}}
			-oft -svfsingle -revd -op "FLASH Erase,Program,Verify"
			-if {{name}}_impl/{{name}}_impl.jed -of {{name}}_flash.svf
		''',
		r'''
		{{invoke_tool("ddtcmd")}}
			-oft -svfsingle -revd -op "SRAM Fast Program"
			-if {{name}}_impl/{{name}}_impl.bit -of {{name}}_sram.svf
		''',
	]
	# These numbers were extracted from
	# "MachXO2 sysCLOCK PLL Design and Usage Guide"
	_supported_osch_freqs = [
		2.08, 2.15, 2.22, 2.29, 2.38, 2.46, 2.56, 2.66, 2.77, 2.89,
		3.02, 3.17, 3.33, 3.50, 3.69, 3.91, 4.16, 4.29, 4.43, 4.59,
		4.75, 4.93, 5.12, 5.32, 5.54, 5.78, 6.05, 6.33, 6.65, 7.00,
		7.39, 7.82, 8.31, 8.58, 8.87, 9.17, 9.50, 9.85, 10.23, 10.64,
		11.08, 11.57, 12.09, 12.67, 13.30, 14.00, 14.78, 15.65, 15.65, 16.63,
		17.73, 19.00, 20.46, 22.17, 24.18, 26.60, 29.56, 33.25, 38.00, 44.33,
		53.20, 66.50, 88.67, 133.00
	]

	@property
	def default_clk_constraint(self) -> Clock:
		# Internal high-speed oscillator on MachXO2/MachXO3L devices.
		# It can have a range of frequencies.
		if self.default_clk == 'OSCH':
			assert self.osch_frequency in self._supported_osch_freqs
			return Clock(int(self.osch_frequency * 1e6))
		# Otherwise, use the defined Clock resource.
		return super().default_clk_constraint

	def create_missing_domain(self, name: str) -> Module:
		# Lattice MachXO2/MachXO3L devices have two global set/reset signals: PUR, which is driven at
		# startup by the configuration logic and unconditionally resets every storage element,
		# and GSR, which is driven by user logic and each storage element may be configured as
		# affected or unaffected by GSR. PUR is purely asynchronous, so even though it is
		# a low-skew global network, its de-assertion may violate a setup/hold constraint with
		# relation to a user clock. To avoid this, a GSR/SGSR instance should be driven
		# synchronized to user clock.
		if name == 'sync' and self.default_clk is not None:
			using_osch = False
			if self.default_clk == 'OSCH':
				using_osch = True
				clk_i = Signal()
			else:
				clk_i = self.request(self.default_clk).i
			if self.default_rst is not None:
				rst_i = self.request(self.default_rst).i
			else:
				rst_i = Const(0)

			gsr0 = Signal()
			gsr1 = Signal()
			m = Module()
			# There is no end-of-startup signal on MachXO2/MachXO3L, but PUR is released after IOB
			# enable, so a simple reset synchronizer (with PUR as the asynchronous reset) does the job.
			m.submodules += [
				Instance('FD1S3AX', p_GSR = 'DISABLED', i_CK = clk_i, i_D = ~rst_i, o_Q = gsr0),
				Instance('FD1S3AX', p_GSR = 'DISABLED', i_CK = clk_i, i_D = gsr0,   o_Q = gsr1),
				# Although we already synchronize the reset input to user clock, SGSR has dedicated
				# clock routing to the center of the FPGA; use that just in case it turns out to be
				# more reliable. (None of this is documented.)
				Instance('SGSR', i_CLK = clk_i, i_GSR = gsr1),
			]
			if using_osch:
				osch_freq = self.osch_frequency
				if osch_freq not in self._supported_osch_freqs:
					raise ValueError(
						f'Frequency {osch_freq!r} is not valid for OSCH clock. Valid frequencies are '
						f'{self._supported_osch_freqs!r}'
					)
				osch_freq_param = f'{float(osch_freq):.2f}'
				m.submodules += [
					Instance(
						'OSCH',
						p_NOM_FREQ = osch_freq_param,
						i_STDBY = Const(0),
						o_OSC = clk_i,
						o_SEDSTDBY = Signal()
					)
				]
			# GSR implicitly connects to every appropriate storage element. As such, the sync
			# domain is reset-less; domains driven by other clocks would need to have dedicated
			# reset circuitry or otherwise meet setup/hold constraints on their own.
			m.domains += ClockDomain('sync', reset_less = True)
			m.d.comb += ClockSignal('sync').eq(clk_i)
			return m

	_single_ended_io_types = [
		'PCI33', 'LVTTL33', 'LVCMOS33', 'LVCMOS25', 'LVCMOS18', 'LVCMOS15', 'LVCMOS12',
		'LVCMOS25R33', 'LVCMOS18R33', 'LVCMOS18R25', 'LVCMOS15R33', 'LVCMOS15R25', 'LVCMOS12R33',
		'LVCMOS12R25', 'LVCMOS10R33', 'LVCMOS10R25', 'SSTL25_I', 'SSTL25_II', 'SSTL18_I',
		'SSTL18_II', 'HSTL18_I', 'HSTL18_II',
	]
	_differential_io_types = [
		'LVDS25', 'LVDS25E', 'RSDS25', 'RSDS25E', 'BLVDS25', 'BLVDS25E', 'MLVDS25', 'MLVDS25E',
		'LVPECL33', 'LVPECL33E', 'SSTL25D_I', 'SSTL25D_II', 'SSTL18D_I', 'SSTL18D_II',
		'HSTL18D_I', 'HSTL18D_II', 'LVTTL33D', 'LVCMOS33D', 'LVCMOS25D', 'LVCMOS18D', 'LVCMOS15D',
		'LVCMOS12D', 'MIPI',
	]

	def should_skip_port_component(
		self, port: Subsignal, attrs: Attrs, component: Literal['io', 'i', 'o', 'p', 'n', 'oe']
	) -> bool:
		# On ECP5, a differential IO is placed by only instantiating an IO buffer primitive at
		# the PIOA or PIOC location, which is always the non-inverting pin.
		if attrs.get('IO_TYPE', 'LVCMOS25') in self._differential_io_types and component == 'n':
			return True
		return False

	def _get_xdr_buffer(
		self, m: Module, pin: Pin, *, i_invert: bool = False, o_invert: bool = False
	) -> tuple[Optional[Signal], Optional[Signal], Optional[Signal]]:
		def get_ireg(clk: Signal, d: Signal, q: Signal) -> None:
			for bit in range(len(q)):
				m.submodules += Instance(
					'IFS1P3DX',
					i_SCLK = clk,
					i_SP = Const(1),
					i_CD = Const(0),
					i_D = d[bit],
					o_Q = q[bit]
				)

		def get_oreg(clk: Signal, d: Signal, q: Signal) -> None:
			for bit in range(len(q)):
				m.submodules += Instance(
					'OFS1P3DX',
					i_SCLK = clk,
					i_SP = Const(1),
					i_CD = Const(0),
					i_D = d[bit],
					o_Q = q[bit]
				)

		def get_iddr(sclk: Signal, d: Signal, q0: Signal, q1: Signal) -> None:
			for bit in range(len(d)):
				m.submodules += Instance(
					'IDDRXE',
					i_SCLK = sclk,
					i_RST = Const(0),
					i_D = d[bit],
					o_Q0 = q0[bit], o_Q1 = q1[bit]
				)

		def get_oddr(sclk: Signal, d0: Signal, d1: Signal, q: Signal) -> None:
			for bit in range(len(q)):
				m.submodules += Instance(
					'ODDRXE',
					i_SCLK = sclk,
					i_RST = Const(0),
					i_D0 = d0[bit], i_D1 = d1[bit],
					o_Q = q[bit]
				)

		def get_ineg(z: Signal, invert: bool) -> Signal:
			if invert:
				a = Signal.like(z, name_suffix = '_n')
				m.d.comb += z.eq(~a)
				return a
			else:
				return z

		def get_oneg(a: Signal, invert: bool) -> Signal:
			if invert:
				z = Signal.like(a, name_suffix = '_n')
				m.d.comb += z.eq(~a)
				return z
			else:
				return a

		if 'i' in pin.dir:
			if pin.xdr < 2:
				pin_i  = get_ineg(pin.i,  i_invert)
			elif pin.xdr == 2:
				pin_i0 = get_ineg(pin.i0, i_invert)
				pin_i1 = get_ineg(pin.i1, i_invert)
		if 'o' in pin.dir:
			if pin.xdr < 2:
				pin_o  = get_oneg(pin.o,  o_invert)
			elif pin.xdr == 2:
				pin_o0 = get_oneg(pin.o0, o_invert)
				pin_o1 = get_oneg(pin.o1, o_invert)

		i = o = t = None
		if 'i' in pin.dir:
			i = Signal(pin.width, name = f'{pin.name}_xdr_i')
		if 'o' in pin.dir:
			o = Signal(pin.width, name = f'{pin.name}_xdr_o')
		if pin.dir in ('oe', 'io'):
			t = Signal(1,         name = f'{pin.name}_xdr_t')

		if pin.xdr == 0:
			if 'i' in pin.dir:
				i = pin_i
			if 'o' in pin.dir:
				o = pin_o
			if pin.dir in ('oe', 'io'):
				t = ~pin.oe
		elif pin.xdr == 1:
			# Note that currently nextpnr will not pack an FF (*FS1P3DX) into the PIO.
			if 'i' in pin.dir:
				get_ireg(pin.i_clk, i, pin_i)
			if 'o' in pin.dir:
				get_oreg(pin.o_clk, pin_o, o)
			if pin.dir in ('oe', 'io'):
				get_oreg(pin.o_clk, ~pin.oe, t)
		elif pin.xdr == 2:
			if 'i' in pin.dir:
				get_iddr(pin.i_clk, i, pin_i0, pin_i1)
			if 'o' in pin.dir:
				get_oddr(pin.o_clk, pin_o0, pin_o1, o)
			if pin.dir in ('oe', 'io'):
				# It looks like Diamond will not pack an OREG as a tristate register in a DDR PIO.
				# It is not clear what is the recommended set of primitives for this task.
				# Similarly, nextpnr will not pack anything as a tristate register in a DDR PIO.
				get_oreg(pin.o_clk, ~pin.oe, t)
		else:
			assert False

		return (i, o, t)

	def get_input(self, pin: Pin, port: Record, attrs: Attrs, invert: bool) -> Module:
		self._check_feature(
			'single-ended input', pin, attrs, valid_xdrs = (0, 1, 2), valid_attrs = True
		)

		m = Module()
		i, o, t = self._get_xdr_buffer(m, pin, i_invert = invert)
		for bit in range(len(port)):
			m.submodules[f'{pin.name}_{bit}'] = Instance(
				'IB',
				i_I = port.io[bit],
				o_O = i[bit]
			)
		return m

	def get_output(self, pin: Pin, port: Record, attrs: Attrs, invert: bool) -> Module:
		self._check_feature(
			'single-ended output', pin, attrs, valid_xdrs = (0, 1, 2), valid_attrs = True
		)

		m = Module()
		i, o, t = self._get_xdr_buffer(m, pin, o_invert = invert)
		for bit in range(len(port)):
			m.submodules[f'{pin.name}_{bit}'] = Instance(
				'OB',
				i_I = o[bit],
				o_O = port.io[bit]
			)
		return m

	def get_tristate(self, pin: Pin, port: Record, attrs: Attrs, invert: bool) -> Module:
		self._check_feature(
			'single-ended tristate', pin, attrs, valid_xdrs = (0, 1, 2), valid_attrs = True
		)

		m = Module()
		i, o, t = self._get_xdr_buffer(m, pin, o_invert = invert)
		for bit in range(len(port)):
			m.submodules[f'{pin.name}_{bit}'] = Instance(
				'OBZ',
				i_T = t,
				i_I = o[bit],
				o_O = port.io[bit]
			)
		return m

	def get_input_output(self, pin: Pin, port: Record, attrs: Attrs, invert: bool) -> Module:
		self._check_feature(
			'single-ended input/output', pin, attrs, valid_xdrs = (0, 1, 2), valid_attrs = True
		)

		m = Module()
		i, o, t = self._get_xdr_buffer(m, pin, i_invert = invert, o_invert = invert)
		for bit in range(len(port)):
			m.submodules[f'{pin.name}_{bit}'] = Instance(
				'BB',
				i_T = t,
				i_I = o[bit],
				o_O = i[bit],
				io_B = port.io[bit]
			)
		return m

	def get_diff_input(self, pin: Pin, port: Record, attrs: Attrs, invert: bool) -> Module:
		self._check_feature(
			'differential input', pin, attrs, valid_xdrs = (0, 1, 2), valid_attrs = True
		)

		m = Module()
		i, o, t = self._get_xdr_buffer(m, pin, i_invert = invert)
		for bit in range(pin.width):
			m.submodules[f'{pin.name}_{bit}'] = Instance(
				'IB',
				i_I = port.p[bit],
				o_O = i[bit]
			)
		return m

	def get_diff_output(self, pin: Pin, port: Record, attrs: Attrs, invert: bool) -> Module:
		self._check_feature(
			'differential output', pin, attrs, valid_xdrs = (0, 1, 2), valid_attrs = True
		)

		m = Module()
		i, o, t = self._get_xdr_buffer(m, pin, o_invert = invert)
		for bit in range(pin.width):
			m.submodules[f'{pin.name}_{bit}'] = Instance(
				'OB',
				i_I = o[bit],
				o_O = port.p[bit],
			)
		return m

	def get_diff_tristate(self, pin: Pin, port: Record, attrs: Attrs, invert: bool) -> Module:
		self._check_feature(
			'differential tristate', pin, attrs, valid_xdrs = (0, 1, 2), valid_attrs = True
		)

		m = Module()
		i, o, t = self._get_xdr_buffer(m, pin, o_invert = invert)
		for bit in range(pin.width):
			m.submodules[f'{pin.name}_{bit}'] = Instance(
				'OBZ',
				i_T = t,
				i_I = o[bit],
				o_O = port.p[bit],
			)
		return m

	def get_diff_input_output(self, pin: Pin, port: Record, attrs: Attrs, invert: bool) -> Module:
		self._check_feature(
			'differential input/output', pin, attrs, valid_xdrs = (0, 1, 2), valid_attrs = True
		)

		m = Module()
		i, o, t = self._get_xdr_buffer(m, pin, i_invert = invert, o_invert = invert)
		for bit in range(pin.width):
			m.submodules[f'{pin.name}_{bit}'] = Instance(
				'BB',
				i_T = t,
				i_I = o[bit],
				o_O = i[bit],
				io_B = port.p[bit],
			)
		return m

	# CDC primitives are not currently specialized for MachXO2/MachXO3L.


MachXO2Platform = MachXO2Or3LPlatform
MachXO3LPlatform = MachXO2Or3LPlatform



# SPDX-License-Identifier: BSD-2-Clause
# torii: UnusedElaboratable=no

from torii                  import *
from torii.lib.fifo         import SyncFIFO
from torii.lib.io           import pin_layout
from torii.sim              import *

from torii.lib.stdio.serial import *

from ...utils               import ToriiTestSuiteCase

def simulation_test(dut, process):
	sim = Simulator(dut)
	with sim.write_vcd('test.vcd'):
		sim.add_clock(1e-6)
		sim.add_sync_process(process)
		sim.run()


class AsyncSerialRXTestCase(ToriiTestSuiteCase):
	def tx_period(self):
		for _ in range((yield self.dut.divisor)):
			yield

	def tx_bits(self, bits):
		for bit in bits:
			yield from self.tx_period()
			yield self.dut.i.eq(bit)

	def rx_test(self, bits, *, data = None, errors = None):
		def process():
			self.assertFalse((yield self.dut.rdy))
			yield self.dut.ack.eq(1)
			yield from self.tx_bits(bits)
			while not (yield self.dut.rdy):
				yield
			if data is not None:
				self.assertFalse((yield self.dut.err.overflow))
				self.assertFalse((yield self.dut.err.frame))
				self.assertFalse((yield self.dut.err.parity))
				self.assertEqual((yield self.dut.data), data)
			if errors is not None:
				for error in errors:
					self.assertTrue((yield getattr(self.dut.err, error)))
		simulation_test(self.dut, process)

	def test_invalid_divisor(self):
		with self.assertRaisesRegex(
			ValueError,
			r'Invalid divisor 1; must be greater than or equal to 5'
		):
			self.dut = AsyncSerialRX(divisor = 1)

	def test_invalid_parity(self):
		with self.assertRaisesRegex(
			ValueError,
			r'Invalid parity \'bad\'; must be one of none, mark, space, even, odd'
		):
			self.dut = AsyncSerialRX(divisor = 5, parity = 'bad')

	def test_8n1(self):
		self.dut = AsyncSerialRX(divisor = 5, data_bits = 8, parity = 'none')
		self.rx_test([0, 1, 0, 1, 0, 1, 1, 1, 0, 1], data = 0b01110101)

	def test_16n1(self):
		self.dut = AsyncSerialRX(divisor = 5, data_bits = 16, parity = 'none')
		self.rx_test(
			[0, 1, 0, 1, 0, 1, 1, 1, 0, 1, 1, 1, 1, 0, 0, 0, 0, 1],
			data = 0b0000111101110101
		)

	def test_8m1(self):
		self.dut = AsyncSerialRX(divisor = 5, data_bits = 8, parity = 'mark')
		self.rx_test([0, 1, 0, 1, 0, 1, 1, 1, 0, 1, 1], data = 0b01110101)
		self.rx_test([0, 1, 0, 1, 0, 1, 1, 0, 0, 1, 1], data = 0b00110101)
		self.rx_test([0, 1, 0, 1, 0, 1, 1, 1, 0, 0, 1], errors = {'parity'})

	def test_8s1(self):
		self.dut = AsyncSerialRX(divisor = 5, data_bits = 8, parity = 'space')
		self.rx_test([0, 1, 0, 1, 0, 1, 1, 1, 0, 0, 1], data = 0b01110101)
		self.rx_test([0, 1, 0, 1, 0, 1, 1, 0, 0, 0, 1], data = 0b00110101)
		self.rx_test([0, 1, 0, 1, 0, 1, 1, 1, 0, 1, 1], errors = {'parity'})

	def test_8e1(self):
		self.dut = AsyncSerialRX(divisor = 5, data_bits = 8, parity = 'even')
		self.rx_test([0, 1, 0, 1, 0, 1, 1, 1, 0, 1, 1], data = 0b01110101)
		self.rx_test([0, 1, 0, 1, 0, 1, 1, 0, 0, 0, 1], data = 0b00110101)
		self.rx_test([0, 1, 0, 1, 0, 1, 1, 1, 0, 0, 1], errors = {'parity'})

	def test_8o1(self):
		self.dut = AsyncSerialRX(divisor = 5, data_bits = 8, parity = 'odd')
		self.rx_test([0, 1, 0, 1, 0, 1, 1, 1, 0, 0, 1], data = 0b01110101)
		self.rx_test([0, 1, 0, 1, 0, 1, 1, 0, 0, 1, 1], data = 0b00110101)
		self.rx_test([0, 1, 0, 1, 0, 1, 1, 1, 0, 1, 1], errors = {'parity'})

	def test_err_frame(self):
		self.dut = AsyncSerialRX(divisor = 5)
		self.rx_test([0, 0, 0, 0, 0, 0, 0, 0, 0, 0], errors = {'frame'})

	def test_err_overflow(self):
		self.dut = AsyncSerialRX(divisor = 5)

		def process():
			self.assertFalse((yield self.dut.rdy))
			yield from self.tx_bits([0, 0, 0, 0, 0, 0, 0, 0, 0, 1])
			yield from self.tx_period()
			yield
			self.assertFalse((yield self.dut.rdy))
			self.assertTrue((yield self.dut.err.overflow))

		simulation_test(self.dut, process)

	def test_fifo(self):
		self.dut  = AsyncSerialRX(divisor = 5)
		self.fifo = SyncFIFO(width = 8, depth = 4)
		m = Module()
		m.submodules.rx   = self.dut
		m.submodules.fifo = self.fifo
		m.d.comb += [
			self.dut.ack.eq(self.fifo.w_rdy),
			self.fifo.w_en.eq(self.dut.rdy),
			self.fifo.w_data.eq(self.dut.data),
		]

		def process():
			yield from self.tx_bits([
				0, 1, 0, 1, 0, 1, 0, 1, 0, 1,
				0, 0, 1, 0, 1, 0, 1, 0, 1, 1
			])
			self.assertTrue((yield self.fifo.r_rdy))
			self.assertEqual((yield self.fifo.r_data), 0x55)
			yield self.fifo.r_en.eq(1)
			yield
			yield
			while not (yield self.fifo.r_rdy):
				yield
			self.assertEqual((yield self.fifo.r_data), 0xAA)
			yield
			self.assertFalse((yield self.fifo.r_rdy))

		simulation_test(m, process)


class AsyncSerialTXTestCase(ToriiTestSuiteCase):
	def tx_period(self):
		for _ in range((yield self.dut.divisor)):
			yield

	def tx_test(self, data, *, bits):
		def process():
			self.assertTrue((yield self.dut.rdy))
			yield self.dut.data.eq(data)
			yield self.dut.ack.eq(1)
			while (yield self.dut.rdy):
				yield
			for bit in bits:
				yield from self.tx_period()
				self.assertEqual((yield self.dut.o), bit)

		simulation_test(self.dut, process)

	def test_invalid_divisor(self):
		with self.assertRaisesRegex(
			ValueError,
			r'Invalid divisor 0; must be greater than or equal to 1'
		):
			self.dut = AsyncSerialTX(divisor = 0)

	def test_invalid_parity(self):
		with self.assertRaisesRegex(
			ValueError,
			r'Invalid parity \'bad\'; must be one of none, mark, space, even, odd'
		):
			self.dut = AsyncSerialTX(divisor = 1, parity = 'bad')

	def test_8n1(self):
		self.dut = AsyncSerialTX(divisor = 1, data_bits = 8, parity = 'none')
		self.tx_test(0b01110101, bits = [0, 1, 0, 1, 0, 1, 1, 1, 0, 1])

	def test_16n1(self):
		self.dut = AsyncSerialTX(divisor = 1, data_bits = 16, parity = 'none')
		self.tx_test(0b0000111101110101, bits = [
			0, 1, 0, 1, 0, 1, 1, 1, 0, 1, 1, 1, 1, 0, 0, 0, 0, 1
		])

	def test_8m1(self):
		self.dut = AsyncSerialTX(divisor = 1, data_bits = 8, parity = 'mark')
		self.tx_test(0b01110101, bits = [0, 1, 0, 1, 0, 1, 1, 1, 0, 1, 1])
		self.tx_test(0b00110101, bits = [0, 1, 0, 1, 0, 1, 1, 0, 0, 1, 1])

	def test_8s1(self):
		self.dut = AsyncSerialTX(divisor = 1, data_bits = 8, parity = 'space')
		self.tx_test(0b01110101, bits = [0, 1, 0, 1, 0, 1, 1, 1, 0, 0, 1])
		self.tx_test(0b00110101, bits = [0, 1, 0, 1, 0, 1, 1, 0, 0, 0, 1])

	def test_8e1(self):
		self.dut = AsyncSerialTX(divisor = 1, data_bits = 8, parity = 'even')
		self.tx_test(0b01110101, bits = [0, 1, 0, 1, 0, 1, 1, 1, 0, 1, 1])
		self.tx_test(0b00110101, bits = [0, 1, 0, 1, 0, 1, 1, 0, 0, 0, 1])

	def test_8o1(self):
		self.dut = AsyncSerialTX(divisor = 1, data_bits = 8, parity = 'odd')
		self.tx_test(0b01110101, bits = [0, 1, 0, 1, 0, 1, 1, 1, 0, 0, 1])
		self.tx_test(0b00110101, bits = [0, 1, 0, 1, 0, 1, 1, 0, 0, 1, 1])

	def test_fifo(self):
		self.dut  = AsyncSerialTX(divisor = 1)
		self.fifo = SyncFIFO(width = 8, depth = 4)
		m = Module()
		m.submodules.tx   = self.dut
		m.submodules.fifo = self.fifo
		m.d.comb += [
			self.dut.ack.eq(self.fifo.r_rdy),
			self.dut.data.eq(self.fifo.r_data),
			self.fifo.r_en.eq(self.fifo.r_rdy & self.dut.rdy),
		]

		def process():
			self.assertTrue((yield self.fifo.w_rdy))
			yield self.fifo.w_en.eq(1)
			yield self.fifo.w_data.eq(0x55)
			yield
			self.assertTrue((yield self.fifo.w_rdy))
			yield self.fifo.w_data.eq(0xAA)
			yield
			yield self.fifo.w_en.eq(0)
			yield
			for bit in [0, 1, 0, 1, 0, 1, 0, 1, 0, 1]:
				yield from self.tx_period()
				self.assertEqual((yield self.dut.o), bit)
			yield
			for bit in [0, 0, 1, 0, 1, 0, 1, 0, 1, 1]:
				yield from self.tx_period()
				self.assertEqual((yield self.dut.o), bit)

		simulation_test(m, process)


class AsyncSerialTestCase(ToriiTestSuiteCase):
	def test_loopback(self):
		pins = Record([
			('rx', pin_layout(1, dir = 'i')),
			('tx', pin_layout(1, dir = 'o'))
		])
		self.dut = AsyncSerial(divisor = 5, pins = pins)
		m = Module()
		m.submodules.serial = self.dut
		m.d.comb += pins.rx.i.eq(pins.tx.o)

		def process():
			self.assertTrue((yield self.dut.tx.rdy))
			yield self.dut.tx.data.eq(0xAA)
			yield self.dut.tx.ack.eq(1)
			yield
			yield self.dut.tx.ack.eq(0)
			yield self.dut.rx.ack.eq(1)
			yield
			while not (yield self.dut.rx.rdy):
				yield
			self.assertEqual((yield self.dut.rx.data), 0xAA)

		simulation_test(m, process)



# SPDX-License-Identifier: BSD-2-Clause

from torii            import *
from torii.build.plat import *

from ..utils          import ToriiTestSuiteCase


class MockPlatform(Platform):
	resources  = []
	connectors = []

	required_tools = []

	def toolchain_prepare(self, fragment, name, **kwargs):
		raise NotImplementedError


class PlatformTestCase(ToriiTestSuiteCase):
	def setUp(self):
		self.platform = MockPlatform()

	def test_add_file_str(self):
		self.platform.add_file('x.txt', 'foo')
		self.assertEqual(self.platform.extra_files['x.txt'], 'foo')

	def test_add_file_bytes(self):
		self.platform.add_file('x.txt', b'foo')
		self.assertEqual(self.platform.extra_files['x.txt'], b'foo')

	def test_add_file_exact_duplicate(self):
		self.platform.add_file('x.txt', b'foo')
		self.platform.add_file('x.txt', b'foo')

	def test_add_file_io(self):
		with open(__file__) as f:
			self.platform.add_file('x.txt', f)
		with open(__file__) as f:
			self.assertEqual(self.platform.extra_files['x.txt'], f.read())

	def test_add_file_wrong_filename(self):
		with self.assertRaisesRegex(
			TypeError,
			r'^File name must be a string, not 1$'
		):
			self.platform.add_file(1, '')

	def test_add_file_wrong_contents(self):
		with self.assertRaisesRegex(
			TypeError,
			r'^File contents must be str, bytes, or a file-like object, not 1$'
		):
			self.platform.add_file('foo', 1)

	def test_add_file_wrong_duplicate(self):
		self.platform.add_file('foo', '')
		with self.assertRaisesRegex(ValueError, r'^File \'foo\' already exists$'):
			self.platform.add_file('foo', 'bar')

	def test_iter_files(self):
		self.platform.add_file('foo.v', '')
		self.platform.add_file('bar.v', '')
		self.platform.add_file('baz.vhd', '')
		self.assertEqual(
			list(self.platform.iter_files('.v')),
			['foo.v', 'bar.v']
		)
		self.assertEqual(
			list(self.platform.iter_files('.vhd')),
			['baz.vhd']
		)
		self.assertEqual(
			list(self.platform.iter_files('.v', '.vhd')),
			['foo.v', 'bar.v', 'baz.vhd']
		)



# SPDX-License-Identifier: BSD-2-Clause

from typing    import Literal, Optional

from ..        import Record
from ..hdl.rec import Layout

__all__ = (
	'pin_layout',
	'Pin',
)


def pin_layout(width: int, dir: Literal['i', 'o', 'oe', 'io'], xdr: int = 0) -> Layout:
	'''
	Layout of the platform interface of a pin or several pins, which may be used inside
	user-defined records.

	See :class:`Pin` for details.
	'''
	if not isinstance(width, int) or width < 1:
		raise TypeError(f'Width must be a positive integer, not {width!r}')
	if dir not in ('i', 'o', 'oe', 'io'):
		raise TypeError(f'Direction must be one of \'i\', \'o\', \'io\', or \'oe\', not {dir!r}')
	if not isinstance(xdr, int) or xdr < 0:
		raise TypeError(f'Gearing ratio must be a non-negative integer, not {xdr!r}')

	fields = []
	if dir in ('i', 'io'):
		if xdr > 0:
			fields.append(('i_clk', 1))
		if xdr > 2:
			fields.append(('i_fclk', 1))
		if xdr in (0, 1):
			fields.append(('i', width))
		else:
			for n in range(xdr):
				fields.append((f'i{n}', width))
	if dir in ('o', 'oe', 'io'):
		if xdr > 0:
			fields.append(('o_clk', 1))
		if xdr > 2:
			fields.append(('o_fclk', 1))
		if xdr in (0, 1):
			fields.append(('o', width))
		else:
			for n in range(xdr):
				fields.append((f'o{n}', width))
	if dir in ('oe', 'io'):
		fields.append(('oe', 1))
	return Layout(fields)


class Pin(Record):
	'''
	An interface to an I/O buffer or a group of them that provides uniform access to input, output,
	or tristate buffers that may include a 1:n gearbox. (A 1:2 gearbox is typically called 'DDR'.)

	A :class:`Pin` is identical to a :class:`Record` that uses the corresponding :meth:`pin_layout`
	except that it allows accessing the parameters like ``width`` as attributes. It is legal to use
	a plain :class:`Record` anywhere a :class:`Pin` is used, provided that these attributes are
	not necessary.

	Parameters
	----------
	width : int
		Width of the ``i``/``iN`` and ``o``/``oN`` signals.
	dir : ``'i'``, ``'o'``, ``'io'``, ``'oe'``
		Direction of the buffers. If ``'i'`` is specified, only the ``i``/``iN`` signals are
		present. If ``'o'`` is specified, only the ``o``/``oN`` signals are present. If ``'oe'`` is
		specified, the ``o``/``oN`` signals are present, and an ``oe`` signal is present.
		If ``'io'`` is specified, both the ``i``/``iN`` and ``o``/``oN`` signals are present, and
		an ``oe`` signal is present.
	xdr : int
		Gearbox ratio. If equal to 0, the I/O buffer is combinatorial, and only ``i``/``o``
		signals are present. If equal to 1, the I/O buffer is SDR, and only ``i``/``o`` signals are
		present. If greater than 1, the I/O buffer includes a gearbox, and ``iN``/``oN`` signals
		are present instead, where ``N in range(0, N)``. For example, if ``xdr = 2``, the I/O buffer
		is DDR; the signal ``i0`` reflects the value at the rising edge, and the signal ``i1``
		reflects the value at the falling edge.
	name : str
		Name of the underlying record.

	Attributes
	----------
	i_clk:
		I/O buffer input clock. Synchronizes `i*`. Present if ``xdr`` is nonzero.
	i_fclk:
		I/O buffer input fast clock. Synchronizes `i*` on higher gearbox ratios. Present if ``xdr``
		is greater than 2.
	i : Signal, out
		I/O buffer input, without gearing. Present if ``dir = 'i'`` or ``dir = 'io'``, and ``xdr`` is
		equal to 0 or 1.
	i0, i1, ... : Signal, out
		I/O buffer inputs, with gearing. Present if ``dir = 'i'`` or ``dir = 'io'``, and ``xdr`` is
		greater than 1.
	o_clk:
		I/O buffer output clock. Synchronizes `o*`, including `oe`. Present if ``xdr`` is nonzero.
	o_fclk:
		I/O buffer output fast clock. Synchronizes `o*` on higher gearbox ratios. Present if
		``xdr`` is greater than 2.
	o : Signal, in
		I/O buffer output, without gearing. Present if ``dir = 'o'`` or ``dir = 'io'``, and ``xdr`` is
		equal to 0 or 1.
	o0, o1, ... : Signal, in
		I/O buffer outputs, with gearing. Present if ``dir = 'o'`` or ``dir = 'io'``, and ``xdr`` is
		greater than 1.
	oe : Signal, in
		I/O buffer output enable. Present if ``dir = 'io'`` or ``dir = 'oe'``. Buffers generally
		cannot change direction more than once per cycle, so at most one output enable signal
		is present.
	'''
	def __init__(
		self, width: int, dir: Literal['i', 'o', 'oe', 'io'], *,
		xdr: int = 0, name: Optional[str] = None, src_loc_at: int = 0
	) -> None:
		self.width = width
		self.dir   = dir
		self.xdr   = xdr

		super().__init__(
			pin_layout(self.width, self.dir, self.xdr),
			name = name, src_loc_at = src_loc_at + 1
		)



# SPDX-License-Identifier: BSD-2-Clause

from torii.hdl.cd import *

from ..utils      import ToriiTestSuiteCase


class ClockDomainTestCase(ToriiTestSuiteCase):
	def test_name(self):
		sync = ClockDomain()
		self.assertEqual(sync.name, 'sync')
		self.assertEqual(sync.clk.name, 'clk')
		self.assertEqual(sync.rst.name, 'rst')
		self.assertEqual(sync.local, False)
		pix = ClockDomain()
		self.assertEqual(pix.name, 'pix')
		self.assertEqual(pix.clk.name, 'pix_clk')
		self.assertEqual(pix.rst.name, 'pix_rst')
		cd_pix = ClockDomain()
		self.assertEqual(cd_pix.name, 'pix')
		dom = [ClockDomain('foo')][0]
		self.assertEqual(dom.name, 'foo')
		with self.assertRaisesRegex(
			ValueError,
			r'^Clock domain name must be specified explicitly$'
		):
			ClockDomain()
		cd_reset = ClockDomain(local = True)
		self.assertEqual(cd_reset.local, True)

	def test_edge(self):
		sync = ClockDomain()
		self.assertEqual(sync.clk_edge, 'pos')
		sync = ClockDomain(clk_edge = 'pos')
		self.assertEqual(sync.clk_edge, 'pos')
		sync = ClockDomain(clk_edge = 'neg')
		self.assertEqual(sync.clk_edge, 'neg')

	def test_edge_wrong(self):
		with self.assertRaisesRegex(
			ValueError,
			r'^Domain clock edge must be one of \'pos\' or \'neg\', not \'xxx\'$'
		):
			ClockDomain('sync', clk_edge = 'xxx')

	def test_with_reset(self):
		pix = ClockDomain()
		self.assertIsNotNone(pix.clk)
		self.assertIsNotNone(pix.rst)
		self.assertFalse(pix.async_reset)

	def test_without_reset(self):
		pix = ClockDomain(reset_less = True)
		self.assertIsNotNone(pix.clk)
		self.assertIsNone(pix.rst)
		self.assertFalse(pix.async_reset)

	def test_async_reset(self):
		pix = ClockDomain(async_reset = True)
		self.assertIsNotNone(pix.clk)
		self.assertIsNotNone(pix.rst)
		self.assertTrue(pix.async_reset)

	def test_rename(self):
		sync = ClockDomain()
		self.assertEqual(sync.name, 'sync')
		self.assertEqual(sync.clk.name, 'clk')
		self.assertEqual(sync.rst.name, 'rst')
		sync.rename('pix')
		self.assertEqual(sync.name, 'pix')
		self.assertEqual(sync.clk.name, 'pix_clk')
		self.assertEqual(sync.rst.name, 'pix_rst')

	def test_rename_reset_less(self):
		sync = ClockDomain(reset_less = True)
		self.assertEqual(sync.name, 'sync')
		self.assertEqual(sync.clk.name, 'clk')
		sync.rename('pix')
		self.assertEqual(sync.name, 'pix')
		self.assertEqual(sync.clk.name, 'pix_clk')

	def test_wrong_name_comb(self):
		with self.assertRaisesRegex(
			ValueError,
			r'^Domain \'comb\' may not be clocked$'
		):
			ClockDomain(name = 'comb')



# SPDX-License-Identifier: BSD-2-Clause

from collections import OrderedDict
from typing      import Generator, Literal, Optional, Union

from ..hdl.ast   import Signal, SignalDict
from ..hdl.rec   import Record
from ..lib.io    import Pin
from .dsl        import Attrs, Connector, DiffPairs, Pins, Resource, Subsignal

__all__ = (
	'ResourceError',
	'ResourceManager',
)


class ResourceError(Exception):
	pass


class ResourceManager:
	def __init__(self, resources: list[Resource], connectors: list[Connector]) -> None:
		self.resources  = OrderedDict()
		self._requested = OrderedDict()
		self._phys_reqd = OrderedDict()

		self.connectors = OrderedDict()
		self._conn_pins = OrderedDict()

		# Constraint lists
		self._ports     = []
		self._clocks    = SignalDict()

		self.add_resources(resources)
		self.add_connectors(connectors)

	def add_resources(self, resources: list[Resource]) -> None:
		for res in resources:
			if not isinstance(res, Resource):
				raise TypeError(f'Object {res!r} is not a Resource')
			if (res.name, res.number) in self.resources:
				raise NameError(f'Trying to add {res!r}, but {self.resources[res.name, res.number]!r} has the same name and number')

			self.resources[res.name, res.number] = res

	def add_connectors(self, connectors: list[Connector]) -> None:
		for conn in connectors:
			if not isinstance(conn, Connector):
				raise TypeError(f'Object {conn!r} is not a Connector')
			if (conn.name, conn.number) in self.connectors:
				raise NameError(f'Trying to add {conn!r}, but {self.connectors[conn.name, conn.number]!r} has the same name and number')

			self.connectors[conn.name, conn.number] = conn

			for conn_pin, plat_pin in conn:
				assert conn_pin not in self._conn_pins
				self._conn_pins[conn_pin] = plat_pin

	def lookup(self, name: str , number: int = 0) -> Resource:
		if (name, number) not in self.resources:
			raise ResourceError(f'Resource {name}#{number} does not exist')

		return self.resources[name, number]

	def request(
		self, name: str, number: int = 0, *,
		dir: Optional[Literal['i', 'o', 'oe', 'io', '-']] = None,
		xdr: Optional[dict[str, int]] = None
	) -> Union[Record, Pin]:
		resource = self.lookup(name, number)
		if (resource.name, resource.number) in self._requested:
			raise ResourceError(f'Resource {name}#{number} has already been requested')

		def merge_options(
			subsignal: Subsignal,
			dir: Optional[Union[Literal['i', 'o', 'oe', 'io', '-'], dict[str, Literal['i', 'o', 'oe', 'io', '-']]]],
			xdr: Optional[Union[int, dict[str, int]]]
		) -> tuple[
			Union[
				Literal['i', 'o', 'oe', 'io', '-'],
				dict[str, Literal['i', 'o', 'oe', 'io', '-']]
			],
			Union[int, dict[str, int]]
		]:
			if isinstance(subsignal.ios[0], Subsignal):
				if dir is None:
					dir = dict()
				if xdr is None:
					xdr = dict()
				if not isinstance(dir, dict):
					raise TypeError(f'Directions must be a dict, not {dir!r}, because {subsignal!r} has subsignals')

				if not isinstance(xdr, dict):
					raise TypeError(f'Data rate must be a dict, not {xdr!r}, because {subsignal!r} has subsignals')

				for sub in subsignal.ios:
					sub_dir = dir.get(sub.name, None)
					sub_xdr = xdr.get(sub.name, None)
					dir[sub.name], xdr[sub.name] = merge_options(sub, sub_dir, sub_xdr)
			else:
				if dir is None:
					dir = subsignal.ios[0].dir
				if xdr is None:
					xdr = 0
				if dir not in ('i', 'o', 'oe', 'io', '-'):
					raise TypeError(
						f'Direction must be one of \'i\', \'o\', \'oe\', \'io\', or \'-\', not {dir!r}'
					)
				if dir != subsignal.ios[0].dir and not (subsignal.ios[0].dir == 'io' or dir == '-'):
					raise ValueError(
						f'Direction of {subsignal.ios[0]!r} cannot be changed from \'{subsignal.ios[0].dir}\' '
						f'to \'{dir}\'; direction can be changed from \'io\' to \'i\', \'o\', or '
						'\'oe\', or from anything to \'-\''
					)

				if not isinstance(xdr, int) or xdr < 0:
					raise ValueError(f'Data rate of {subsignal.ios[0]!r} must be a non-negative integer, not {xdr!r}')

			return (dir, xdr)

		def resolve(
			resource: Resource,
			dir: Union[Literal['i', 'o', 'oe', 'io', '-'], dict[str, Literal['i', 'o', 'oe', 'io', '-']]],
			xdr: Union[int, dict[str, int]],
			name: str, attrs: Attrs
		) -> Union[Record, Pin]:
			for attr_key, attr_value in attrs.items():
				if hasattr(attr_value, '__call__'):
					attr_value = attr_value(self)
					assert attr_value is None or isinstance(attr_value, str)
				if attr_value is None:
					del attrs[attr_key]
				else:
					attrs[attr_key] = attr_value

			if isinstance(resource.ios[0], Subsignal):
				fields = OrderedDict()
				for sub in resource.ios:
					fields[sub.name] = resolve(
						sub, dir[sub.name], xdr[sub.name],
						name = f'{name}__{sub.name}',
						attrs = {**attrs, **sub.attrs}
					)
				return Record([
					(f_name, f.layout) for (f_name, f) in fields.items()
				], fields = fields, name = name)

			elif isinstance(resource.ios[0], (Pins, DiffPairs)):
				phys = resource.ios[0]
				if isinstance(phys, Pins):
					phys_names = phys.names
					port = Record([('io', len(phys))], name = name)
				if isinstance(phys, DiffPairs):
					phys_names = []
					record_fields = []
					if not self.should_skip_port_component(None, attrs, 'p'):
						phys_names += phys.p.names
						record_fields.append(('p', len(phys)))
					if not self.should_skip_port_component(None, attrs, 'n'):
						phys_names += phys.n.names
						record_fields.append(('n', len(phys)))
					port = Record(record_fields, name = name)
				if dir == '-':
					pin = None
				else:
					pin = Pin(len(phys), dir, xdr = xdr, name = name)

				for phys_name in phys_names:
					if phys_name in self._phys_reqd:
						raise ResourceError(
							f'Resource component {name} uses physical pin {phys_name}, but it '
							f'is already used by resource component {self._phys_reqd[phys_name]} that was '
							'requested earlier'
						)

					self._phys_reqd[phys_name] = name

				self._ports.append((resource, pin, port, attrs))

				if pin is not None and resource.clock is not None:
					self.add_clock_constraint(pin.i, resource.clock.frequency)

				return pin if pin is not None else port

			else:
				assert False # :nocov:

		value = resolve(
			resource,
			*merge_options(resource, dir, xdr),
			name = f'{resource.name}_{resource.number}',
			attrs = resource.attrs
		)
		self._requested[resource.name, resource.number] = value
		return value

	def iter_single_ended_pins(self) -> Generator[tuple[
		Pin, Subsignal, Attrs, bool
	], None, None]:
		for res, pin, port, attrs in self._ports:
			if pin is None:
				continue
			if isinstance(res.ios[0], Pins):
				yield (pin, port, attrs, res.ios[0].invert)

	def iter_differential_pins(self) -> Generator[tuple[
		Pin, Subsignal, Attrs, bool
	], None, None]:
		for res, pin, port, attrs in self._ports:
			if pin is None:
				continue
			if isinstance(res.ios[0], DiffPairs):
				yield (pin, port, attrs, res.ios[0].invert)

	def should_skip_port_component(
		self, port: Subsignal, attrs: Attrs, component: Literal['io', 'i', 'o', 'p', 'n', 'oe']
	) -> bool:
		return False

	def iter_ports(self) -> Generator[Signal, None, None]:
		for res, pin, port, attrs in self._ports:
			if isinstance(res.ios[0], Pins):
				if not self.should_skip_port_component(port, attrs, 'io'):
					yield port.io
			elif isinstance(res.ios[0], DiffPairs):
				if not self.should_skip_port_component(port, attrs, 'p'):
					yield port.p
				if not self.should_skip_port_component(port, attrs, 'n'):
					yield port.n
			else:
				assert False

	def iter_port_constraints(self) -> Generator[
		tuple[str, str, Attrs], None, None
	]:
		for res, pin, port, attrs in self._ports:
			if isinstance(res.ios[0], Pins):
				if not self.should_skip_port_component(port, attrs, 'io'):
					yield (port.io.name, res.ios[0].map_names(self._conn_pins, res), attrs)
			elif isinstance(res.ios[0], DiffPairs):
				if not self.should_skip_port_component(port, attrs, 'p'):
					yield (port.p.name, res.ios[0].p.map_names(self._conn_pins, res), attrs)
				if not self.should_skip_port_component(port, attrs, 'n'):
					yield (port.n.name, res.ios[0].n.map_names(self._conn_pins, res), attrs)
			else:
				assert False

	def iter_port_constraints_bits(self) -> Generator[
		tuple[str, str, Attrs], None, None
	]:
		for port_name, pin_names, attrs in self.iter_port_constraints():
			if len(pin_names) == 1:
				yield (port_name, pin_names[0], attrs)
			else:
				for bit, pin_name in enumerate(pin_names):
					yield (f'{port_name}[{bit}]', pin_name, attrs)

	def add_clock_constraint(self, clock: Signal, frequency: Union[int, float]) -> None:
		if not isinstance(clock, Signal):
			raise TypeError(f'Object {clock!r} is not a Signal')
		if not isinstance(frequency, (int, float)):
			raise TypeError(f'Frequency must be a number, not {frequency!r}')

		if clock in self._clocks:
			raise ValueError(f'Cannot add clock constraint on {clock!r}, which is already constrained to {self._clocks[clock]} Hz')

		else:
			self._clocks[clock] = float(frequency)

	def iter_clock_constraints(self) -> Generator[
		tuple[str, Signal, float], None, None
	]:
		# Back-propagate constraints through the input buffer. For clock constraints on pins
		# (the majority of cases), toolchains work better if the constraint is defined on the pin
		# and not on the buffered internal net; and if the toolchain is advanced enough that
		# it considers clock phase and delay of the input buffer, it is *necessary* to define
		# the constraint on the pin to match the designer's expectation of phase being referenced
		# to the pin.
		#
		# Constraints on nets with no corresponding input pin (e.g. PLL or SERDES outputs) are not
		# affected.
		pin_i_to_port = SignalDict()
		for res, pin, port, attrs in self._ports:
			if hasattr(pin, 'i'):
				if isinstance(res.ios[0], Pins):
					pin_i_to_port[pin.i] = port.io
				elif isinstance(res.ios[0], DiffPairs):
					pin_i_to_port[pin.i] = port.p
				else:
					raise ValueError(f'Expected res.ios[0] to be a \'Pins\' or \'DiffPairs\', not {res.ios[0]!r}')

		for net_signal, frequency in self._clocks.items():
			port_signal = pin_i_to_port.get(net_signal)
			yield (net_signal, port_signal, frequency)



# SPDX-License-Identifier: BSD-2-Clause

from warnings       import warn

from .lattice.ice40 import ICE40Platform as LatticeICE40Platform

__all__ = (
	'LatticeICE40Platform',
)

warn(
	'Instead of torii.platform.vendor.lattice_ice40, use torii.platform.vendor.lattice.ice40',
	DeprecationWarning, stacklevel = 2
)



# SPDX-License-Identifier: BSD-2-Clause

import operator
from collections import OrderedDict
from typing      import Optional, List

from ..util      import tracer
from .ast        import (
	Array, Cat, ClockSignal, Const, Signal, Switch, Shape, Mux,
	unsigned
)
from .ir         import Elaboratable, Instance, Fragment

__all__ = (
	'DummyPort',
	'Memory',
	'ReadPort',
	'WritePort',
)


class Memory(Elaboratable):
	'''
	A word addressable storage.

	Parameters
	----------
	width : int
		Access granularity. Each storage element of this memory is ``width`` bits in size.
	depth : int
		Word count. This memory contains ``depth`` storage elements.
	init : list of int
		Initial values. At power on, each storage element in this memory is initialized to
		the corresponding element of ``init``, if any, or to zero otherwise.
		Uninitialized memories are not currently supported.
	name : str
		Name hint for this memory. If ``None`` (default) the name is inferred from the variable
		name this ``Signal`` is assigned to.
	attrs : dict
		Dictionary of synthesis attributes.

	Attributes
	----------
	width : int
	depth : int
	init : list of int
	attrs : dict

	'''

	def __init__(
		self, *, width: int, depth: int, init = None, name: Optional[str] = None,
		attrs: Optional[OrderedDict] = None, simulate: bool = True
	) -> None:
		if not isinstance(width, int) or width < 0:
			raise TypeError(f'Memory width must be a non-negative integer, not {width!r}')
		if not isinstance(depth, int) or depth < 0:
			raise TypeError(f'Memory depth must be a non-negative integer, not {depth!r}')

		self.name    = name or tracer.get_var_name(depth = 2, default = '$memory')
		self.src_loc = tracer.get_src_loc()

		self.width = width
		self.depth = depth
		self.attrs = OrderedDict(() if attrs is None else attrs)

		# Array of signals for simulation.
		self._array = Array()
		if simulate:
			for addr in range(self.depth):
				self._array.append(Signal(self.width, name = f'{name or "memory"}({addr})'))

		self.init = init
		self._read_ports: List[ReadPort] = []
		self._write_ports: List[WritePort] = []

	@property
	def init(self):
		return self._init

	@init.setter
	def init(self, new_init):
		self._init = [] if new_init is None else list(new_init)
		if len(self.init) > self.depth:
			raise ValueError(f'Memory initialization value count exceed memory depth ({len(self.init)} > {self.depth})')

		try:
			for addr in range(len(self._array)):
				if addr < len(self._init):
					self._array[addr].reset = operator.index(self._init[addr])
				else:
					self._array[addr].reset = 0
		except TypeError as e:
			raise TypeError(f'Memory initialization value at address {addr:x}: {e}') from None

	def read_port(self, *, src_loc_at = 0, **kwargs):
		'''
		Get a read port.

		See :class:`ReadPort` for details.

		Arguments
		---------
		domain : str
		transparent : bool

		Returns
		-------
		An instance of :class:`ReadPort` associated with this memory.

		'''

		return ReadPort(self, src_loc_at = 1 + src_loc_at, **kwargs)

	def write_port(self, *, src_loc_at = 0, **kwargs):
		'''
		Get a write port.

		See :class:`WritePort` for details.

		Arguments
		---------
		domain : str
		granularity : int

		Returns
		-------
		An instance of :class:`WritePort` associated with this memory.

		'''

		return WritePort(self, src_loc_at = 1 + src_loc_at, **kwargs)

	def __getitem__(self, index):
		''' Simulation only. '''
		return self._array[index]

	def elaborate(self, platform):
		init = ''.join(format(Const(elem, unsigned(self.width)).value, f'0{self.width}b') for elem in reversed(self.init))
		init = Const(int(init or '0', 2), self.depth * self.width)
		rd_clk = []
		rd_clk_enable = 0
		rd_transparency_mask = 0
		for index, port in enumerate(self._read_ports):
			if port.domain != 'comb':
				rd_clk.append(ClockSignal(port.domain))
				rd_clk_enable |= 1 << index
				if port.transparent:
					for write_index, write_port in enumerate(self._write_ports):
						if port.domain == write_port.domain:
							rd_transparency_mask |= 1 << (index * len(self._write_ports) + write_index)
			else:
				rd_clk.append(Const(0, 1))

		f = Instance(
			'$mem_v2',
			*(('a', attr, value) for attr, value in self.attrs.items()),
			p_SIZE                 = self.depth,
			p_OFFSET               = 0,
			p_ABITS                = Shape.cast(range(self.depth)).width,
			p_WIDTH                = self.width,
			p_INIT                 = init,
			p_RD_PORTS             = len(self._read_ports),
			p_RD_CLK_ENABLE        = Const(rd_clk_enable, len(self._read_ports)) if self._read_ports else Const(0, 1),
			p_RD_CLK_POLARITY      = Const(-1, unsigned(len(self._read_ports))) if self._read_ports else Const(0, 1),
			p_RD_TRANSPARENCY_MASK = Const(rd_transparency_mask, max(1, len(self._read_ports) * len(self._write_ports))),
			p_RD_COLLISION_X_MASK  = Const(0, max(1, len(self._read_ports) * len(self._write_ports))),
			p_RD_WIDE_CONTINUATION = Const(0, len(self._read_ports)) if self._read_ports else Const(0, 1),
			p_RD_CE_OVER_SRST      = Const(0, len(self._read_ports)) if self._read_ports else Const(0, 1),
			p_RD_ARST_VALUE        = Const(0, len(self._read_ports) * self.width),
			p_RD_SRST_VALUE        = Const(0, len(self._read_ports) * self.width),
			p_RD_INIT_VALUE        = Const(0, len(self._read_ports) * self.width),
			p_WR_PORTS             = len(self._write_ports),
			p_WR_CLK_ENABLE        = Const(-1, unsigned(len(self._write_ports))) if self._write_ports else Const(0, 1),
			p_WR_CLK_POLARITY      = Const(-1, unsigned(len(self._write_ports))) if self._write_ports else Const(0, 1),
			p_WR_PRIORITY_MASK     = Const(0, len(self._write_ports) * len(self._write_ports)) if self._write_ports else Const(0, 1),
			p_WR_WIDE_CONTINUATION = Const(0, len(self._write_ports)) if self._write_ports else Const(0, 1),
			i_RD_CLK               = Cat(rd_clk),
			i_RD_EN                = Cat(port.en for port in self._read_ports),
			i_RD_ARST              = Const(0, len(self._read_ports)),
			i_RD_SRST              = Const(0, len(self._read_ports)),
			i_RD_ADDR              = Cat(port.addr for port in self._read_ports),
			o_RD_DATA              = Cat(port.data for port in self._read_ports),
			i_WR_CLK               = Cat(ClockSignal(port.domain) for port in self._write_ports),
			i_WR_EN                = Cat(
				Cat(
					en_bit.replicate(port.granularity) for en_bit in port.en
				) for port in self._write_ports
			),
			i_WR_ADDR              = Cat(port.addr for port in self._write_ports),
			i_WR_DATA              = Cat(port.data for port in self._write_ports),
		)

		for port in self._read_ports:
			port._MustUse__used = True
			if port.domain == 'comb':
				f.add_statements(port.data.eq(self._array[port.addr]))
				f.add_driver(port.data)
			else:
				data = self._array[port.addr]
				for write_port in self._write_ports:
					if port.domain == write_port.domain and port.transparent:
						if len(write_port.en) > 1:
							parts = []
							for index, en_bit in enumerate(write_port.en):
								offset = index * write_port.granularity
								bits = slice(offset, offset + write_port.granularity)
								cond = en_bit & (port.addr == write_port.addr)
								parts.append(Mux(cond, write_port.data[bits], data[bits]))
							data = Cat(parts)
						else:
							cond = write_port.en & (port.addr == write_port.addr)
							data = Mux(cond, write_port.data, data)
				f.add_statements(
					Switch(port.en, {
						1: port.data.eq(data)
					})
				)
				f.add_driver(port.data, port.domain)
		for port in self._write_ports:
			port._MustUse__used = True
			if len(port.en) > 1:
				for index, en_bit in enumerate(port.en):
					offset = index * port.granularity
					bits   = slice(offset, offset + port.granularity)
					write_data = self._array[port.addr][bits].eq(port.data[bits])
					f.add_statements(Switch(en_bit, { 1: write_data }))
			else:
				write_data = self._array[port.addr].eq(port.data)
				f.add_statements(Switch(port.en, { 1: write_data }))
			for signal in self._array:
				f.add_driver(signal, port.domain)
		return f


class ReadPort(Elaboratable):
	'''
	A memory read port.

	Parameters
	----------
	memory : :class:`Memory`
		Memory associated with the port.
	domain : str
		Clock domain. Defaults to ``'sync'``. If set to ``'comb'``, the port is asynchronous.
		Otherwise, the read data becomes available on the next clock cycle.
	transparent : bool
		Port transparency. If set (default), a read at an address that is also being written to in
		the same clock cycle will output the new value. Otherwise, the old value will be output
		first. This behavior only applies to ports in the same domain.

	Attributes
	----------
	memory : :class:`Memory`
	domain : str
	transparent : bool
	addr : Signal(range(memory.depth)), in
		Read address.
	data : Signal(memory.width), out
		Read data.
	en : Signal or Const, in
		Read enable. If asserted, ``data`` is updated with the word stored at ``addr``.
	Raises
	------
	:class:`ValueError` if the read port is simultaneously asynchronous and non-transparent.

	'''

	def __init__(self, memory, *, domain = 'sync', transparent = True, src_loc_at = 0):
		if domain == 'comb' and not transparent:
			raise ValueError('Read port cannot be simultaneously asynchronous and non-transparent')

		self.memory      = memory
		self.domain      = domain
		self.transparent = transparent

		self.addr = Signal(
			range(memory.depth),
			name = f'{memory.name}_r_addr', src_loc_at = 1 + src_loc_at
		)
		self.data = Signal(
			memory.width,
			name = f'{memory.name}_r_data', src_loc_at = 1 + src_loc_at
		)
		if self.domain != 'comb':
			self.en = Signal(
				name = f'{memory.name}_r_en', reset = 1,
				src_loc_at = 1 + src_loc_at
			)
		else:
			self.en = Const(1)

		memory._read_ports.append(self)

	def elaborate(self, platform):
		if self is self.memory._read_ports[0]:
			return self.memory
		else:
			return Fragment()


class WritePort(Elaboratable):
	'''
	A memory write port.

	Parameters
	----------
	memory : :class:`Memory`
		Memory associated with the port.
	domain : str
		Clock domain. Defaults to ``"sync"``. Writes have a latency of 1 clock cycle.
	granularity : int
		Port granularity. Defaults to ``memory.width``. Write data is split evenly in
		``memory.width // granularity`` chunks, which can be updated independently.

	Attributes
	----------
	memory : :class:`Memory`
	domain : str
	granularity : int
	addr : Signal(range(memory.depth)), in
		Write address.
	data : Signal(memory.width), in
		Write data.
	en : Signal(memory.width // granularity), in
		Write enable. Each bit selects a non-overlapping chunk of ``granularity`` bits on the
		``data`` signal, which is written to memory at ``addr``. Unselected chunks are ignored.

	Raises
	------
	:class:`ValueError` if the write port granularity is greater than memory width, or does not
	divide memory width evenly.

	'''

	def __init__(self, memory, *, domain = 'sync', granularity = None, src_loc_at = 0):
		if granularity is None:
			granularity = memory.width
		if not isinstance(granularity, int) or granularity < 0:
			raise TypeError(f'Write port granularity must be a non-negative integer, not {granularity!r}')
		if granularity > memory.width:
			raise ValueError(f'Write port granularity must not be greater than memory width ({granularity} > {memory.width})')
		if memory.width // granularity * granularity != memory.width:
			raise ValueError('Write port granularity must divide memory width evenly')

		self.memory       = memory
		self.domain       = domain
		self.granularity  = granularity

		self.addr = Signal(
			range(memory.depth),
			name = f'{memory.name}_w_addr', src_loc_at = 1 + src_loc_at
		)
		self.data = Signal(
			memory.width,
			name = f'{memory.name}_w_data', src_loc_at = 1 + src_loc_at
		)
		self.en   = Signal(
			memory.width // granularity,
			name = f'{memory.name}_w_en', src_loc_at = 1 + src_loc_at
		)

		memory._write_ports.append(self)

	def elaborate(self, platform):
		if not self.memory._read_ports and self is self.memory._write_ports[0]:
			return self.memory
		else:
			return Fragment()


class DummyPort:
	'''
	Dummy memory port.

	This port can be used in place of either a read or a write port for testing and verification.
	It does not include any read/write port specific attributes, i.e. none besides ``'domain'``;
	any such attributes may be set manually.

	'''

	def __init__(self, *, data_width, addr_width, domain = 'sync', name = None, granularity = None):
		self.domain = domain

		if granularity is None:
			granularity = data_width
		if name is None:
			name = tracer.get_var_name(depth = 2, default = 'dummy')

		self.addr = Signal(addr_width, name = f'{name}_addr', src_loc_at = 1)
		self.data = Signal(data_width, name = f'{name}_data', src_loc_at = 1)
		self.en   = Signal(data_width // granularity, name = f'{name}_en', src_loc_at = 1)



# SPDX-License-Identifier: BSD-2-Clause
# torii: UnusedElaboratable=no

from torii.hdl      import *
from torii.asserts  import *
from torii.sim      import *
from torii.lib.fifo import *

from ..utils        import ToriiTestSuiteCase


class FIFOTestCase(ToriiTestSuiteCase):
	def test_depth_wrong(self):
		with self.assertRaisesRegex(
			TypeError,
			r'^FIFO width must be a non-negative integer, not -1$'
		):
			FIFOInterface(width = -1, depth = 8, fwft = True)
		with self.assertRaisesRegex(
			TypeError,
			r'^FIFO depth must be a non-negative integer, not -1$'
		):
			FIFOInterface(width = 8, depth = -1, fwft = True)

	def test_sync_depth(self):
		self.assertEqual(SyncFIFO(width = 8, depth = 0).depth, 0)
		self.assertEqual(SyncFIFO(width = 8, depth = 1).depth, 1)
		self.assertEqual(SyncFIFO(width = 8, depth = 2).depth, 2)

	def test_sync_buffered_depth(self):
		self.assertEqual(SyncFIFOBuffered(width = 8, depth = 0).depth, 0)
		self.assertEqual(SyncFIFOBuffered(width = 8, depth = 1).depth, 1)
		self.assertEqual(SyncFIFOBuffered(width = 8, depth = 2).depth, 2)

	def test_async_depth(self):
		self.assertEqual(AsyncFIFO(width = 8, depth = 0 ).depth, 0)
		self.assertEqual(AsyncFIFO(width = 8, depth = 1 ).depth, 1)
		self.assertEqual(AsyncFIFO(width = 8, depth = 2 ).depth, 2)
		self.assertEqual(AsyncFIFO(width = 8, depth = 3 ).depth, 4)
		self.assertEqual(AsyncFIFO(width = 8, depth = 4 ).depth, 4)
		self.assertEqual(AsyncFIFO(width = 8, depth = 15).depth, 16)
		self.assertEqual(AsyncFIFO(width = 8, depth = 16).depth, 16)
		self.assertEqual(AsyncFIFO(width = 8, depth = 17).depth, 32)

	def test_async_depth_wrong(self):
		with self.assertRaisesRegex(
			ValueError, (
				r'^AsyncFIFO only supports depths that are powers of 2; '
				r'requested exact depth 15 is not$'
			)
		):
			AsyncFIFO(width = 8, depth = 15, exact_depth = True)

	def test_async_buffered_depth(self):
		self.assertEqual(AsyncFIFOBuffered(width = 8, depth = 0 ).depth, 0)
		self.assertEqual(AsyncFIFOBuffered(width = 8, depth = 1 ).depth, 2)
		self.assertEqual(AsyncFIFOBuffered(width = 8, depth = 2 ).depth, 2)
		self.assertEqual(AsyncFIFOBuffered(width = 8, depth = 3 ).depth, 3)
		self.assertEqual(AsyncFIFOBuffered(width = 8, depth = 4 ).depth, 5)
		self.assertEqual(AsyncFIFOBuffered(width = 8, depth = 15).depth, 17)
		self.assertEqual(AsyncFIFOBuffered(width = 8, depth = 16).depth, 17)
		self.assertEqual(AsyncFIFOBuffered(width = 8, depth = 17).depth, 17)
		self.assertEqual(AsyncFIFOBuffered(width = 8, depth = 18).depth, 33)

	def test_async_buffered_depth_wrong(self):
		with self.assertRaisesRegex(
			ValueError, (
				r'^AsyncFIFOBuffered only supports depths that are one higher than powers of 2; '
				r'requested exact depth 16 is not$'
			)
		):
			AsyncFIFOBuffered(width = 8, depth = 16, exact_depth = True)

class FIFOModel(Elaboratable, FIFOInterface):
	'''
	Non-synthesizable first-in first-out queue, implemented naively as a chain of registers.
	'''
	def __init__(self, *, width, depth, fwft, r_domain, w_domain):
		super().__init__(width = width, depth = depth, fwft = fwft)

		self.r_domain = r_domain
		self.w_domain = w_domain

		self.level = Signal(range(self.depth + 1))
		self.r_level = Signal(range(self.depth + 1))
		self.w_level = Signal(range(self.depth + 1))

	def elaborate(self, platform):
		m = Module()

		storage = Memory(width = self.width, depth = self.depth)
		w_port  = m.submodules.w_port = storage.write_port(domain = self.w_domain)
		r_port  = m.submodules.r_port = storage.read_port(domain = 'comb')

		produce = Signal(range(self.depth))
		consume = Signal(range(self.depth))

		m.d.comb += self.r_rdy.eq(self.level > 0)
		m.d.comb += r_port.addr.eq((consume + 1) % self.depth)
		if self.fwft:
			m.d.comb += self.r_data.eq(r_port.data)
		with m.If(self.r_en & self.r_rdy):
			if not self.fwft:
				m.d[self.r_domain] += self.r_data.eq(r_port.data)
			m.d[self.r_domain] += consume.eq(r_port.addr)

		m.d.comb += self.w_rdy.eq(self.level < self.depth)
		m.d.comb += w_port.data.eq(self.w_data)
		with m.If(self.w_en & self.w_rdy):
			m.d.comb += w_port.addr.eq((produce + 1) % self.depth)
			m.d.comb += w_port.en.eq(1)
			m.d[self.w_domain] += produce.eq(w_port.addr)

		with m.If(ResetSignal(self.r_domain) | ResetSignal(self.w_domain)):
			m.d.sync += self.level.eq(0)
		with m.Else():
			m.d.sync += self.level.eq(
				self.level + (self.w_rdy & self.w_en) - (self.r_rdy & self.r_en)
			)

		m.d.comb += [
			self.r_level.eq(self.level),
			self.w_level.eq(self.level),
		]
		m.d.comb += Assert(ResetSignal(self.r_domain) == ResetSignal(self.w_domain))

		return m


class FIFOModelEquivalenceSpec(Elaboratable):
	'''
	The first-in first-out queue model equivalence specification: for any inputs and control
	signals, the behavior of the implementation under test exactly matches the ideal model,
	except for behavior not defined by the model.
	'''
	def __init__(self, fifo, r_domain, w_domain):
		self.fifo = fifo

		self.r_domain = r_domain
		self.w_domain = w_domain

	def elaborate(self, platform):
		m = Module()
		m.submodules.dut  = dut  = self.fifo
		m.submodules.gold = gold = FIFOModel(
			width = dut.width, depth = dut.depth, fwft = dut.fwft,
			r_domain = self.r_domain, w_domain = self.w_domain
		)

		m.d.comb += [
			gold.r_en.eq(dut.r_rdy & dut.r_en),
			gold.w_en.eq(dut.w_en),
			gold.w_data.eq(dut.w_data),
		]

		m.d.comb += Assert(dut.r_rdy.implies(gold.r_rdy))
		m.d.comb += Assert(dut.w_rdy.implies(gold.w_rdy))
		m.d.comb += Assert(dut.r_level == gold.r_level)
		m.d.comb += Assert(dut.w_level == gold.w_level)

		if dut.fwft:
			m.d.comb += Assert(
				dut.r_rdy.implies(dut.r_data == gold.r_data)
			)
		else:
			m.d.comb += Assert(
				(
					Past(dut.r_rdy, domain = self.r_domain) &
					Past(dut.r_en, domain = self.r_domain)
				).implies(dut.r_data == gold.r_data)
			)

		return m


class FIFOContractSpec(Elaboratable):
	'''
	The first-in first-out queue contract specification: if two elements are written to the queue
	consecutively, they must be read out consecutively at some later point, no matter all other
	circumstances, with the exception of reset.
	'''
	def __init__(self, fifo, *, r_domain, w_domain, bound):
		self.fifo     = fifo
		self.r_domain = r_domain
		self.w_domain = w_domain
		self.bound    = bound

	def elaborate(self, platform):
		m = Module()
		m.submodules.dut = fifo = self.fifo

		m.domains += ClockDomain('sync')
		m.d.comb += ResetSignal().eq(0)
		if self.w_domain != 'sync':
			m.domains += ClockDomain(self.w_domain)
			m.d.comb += ResetSignal(self.w_domain).eq(0)
		if self.r_domain != 'sync':
			m.domains += ClockDomain(self.r_domain)
			m.d.comb += ResetSignal(self.r_domain).eq(0)

		entry_1 = AnyConst(fifo.width)
		entry_2 = AnyConst(fifo.width)

		with m.FSM(domain = self.w_domain) as write_fsm:
			with m.State('WRITE-1'):
				with m.If(fifo.w_rdy):
					m.d.comb += [
						fifo.w_data.eq(entry_1),
						fifo.w_en.eq(1)
					]
					m.next = 'WRITE-2'
			with m.State('WRITE-2'):
				with m.If(fifo.w_rdy):
					m.d.comb += [
						fifo.w_data.eq(entry_2),
						fifo.w_en.eq(1)
					]
					m.next = 'DONE'
			with m.State('DONE'):
				pass

		with m.FSM(domain = self.r_domain) as read_fsm:
			read_1 = Signal(fifo.width)
			read_2 = Signal(fifo.width)
			with m.State('READ'):
				m.d.comb += fifo.r_en.eq(1)
				if fifo.fwft:
					r_rdy = fifo.r_rdy
				else:
					r_rdy = Past(fifo.r_rdy, domain = self.r_domain)
				with m.If(r_rdy):
					m.d.sync += [
						read_1.eq(read_2),
						read_2.eq(fifo.r_data),
					]
				with m.If((read_1 == entry_1) & (read_2 == entry_2)):
					m.next = 'DONE'
			with m.State('DONE'):
				pass

		with m.If(Initial()):
			m.d.comb += Assume(write_fsm.ongoing('WRITE-1'))
			m.d.comb += Assume(read_fsm.ongoing('READ'))
		with m.If(Past(Initial(), self.bound - 1)):
			m.d.comb += Assert(read_fsm.ongoing('DONE'))

		with m.If(ResetSignal(domain = self.w_domain)):
			m.d.comb += Assert(~fifo.r_rdy)

		if self.w_domain != 'sync' or self.r_domain != 'sync':
			m.d.comb += Assume(
				Rose(ClockSignal(self.w_domain)) |
				Rose(ClockSignal(self.r_domain))
			)

		return m


class FIFOFormalCase(ToriiTestSuiteCase):
	def check_sync_fifo(self, fifo):
		self.assertFormal(
			FIFOModelEquivalenceSpec(fifo, r_domain = 'sync', w_domain = 'sync'),
			mode = 'bmc', depth = fifo.depth + 1
		)
		self.assertFormal(
			FIFOContractSpec(
				fifo, r_domain = 'sync', w_domain = 'sync', bound = fifo.depth * 2 + 1
			),
			mode = 'hybrid', depth = fifo.depth * 2 + 1
		)

	def test_sync_fwft_pot(self):
		self.check_sync_fifo(SyncFIFO(width = 8, depth = 4, fwft = True))

	def test_sync_fwft_npot(self):
		self.check_sync_fifo(SyncFIFO(width = 8, depth = 5, fwft = True))

	def test_sync_not_fwft_pot(self):
		self.check_sync_fifo(SyncFIFO(width = 8, depth = 4, fwft = False))

	def test_sync_not_fwft_npot(self):
		self.check_sync_fifo(SyncFIFO(width = 8, depth = 5, fwft = False))

	def test_sync_buffered_pot(self):
		self.check_sync_fifo(SyncFIFOBuffered(width = 8, depth = 4))

	def test_sync_buffered_potp1(self):
		self.check_sync_fifo(SyncFIFOBuffered(width = 8, depth = 5))

	def test_sync_buffered_potm1(self):
		self.check_sync_fifo(SyncFIFOBuffered(width = 8, depth = 3))

	def check_async_fifo(self, fifo):
		# TODO: properly doing model equivalence checking on this likely requires multiclock,
		# which is not really documented nor is it clear how to use it.
		# self.assertFormal(FIFOModelEquivalenceSpec(fifo, r_domain = 'read', w_domain = 'write'),
		#                   mode = 'bmc', depth = fifo.depth * 3 + 1)
		self.assertFormal(
			FIFOContractSpec(
				fifo, r_domain = 'read', w_domain = 'write', bound = fifo.depth * 4 + 1
			),
			mode = 'hybrid', depth = fifo.depth * 4 + 1
		)

	def test_async(self):
		self.check_async_fifo(AsyncFIFO(width = 8, depth = 4))

	def test_async_buffered(self):
		self.check_async_fifo(AsyncFIFOBuffered(width = 8, depth = 4))


# we need this testcase because we cant do model equivalence checking on the async fifos (at the moment)
class AsyncFIFOSimCase(ToriiTestSuiteCase):
	def test_async_fifo_r_level_latency(self):
		fifo = AsyncFIFO(width = 32, depth = 10, r_domain = 'sync', w_domain = 'sync')

		ff_syncronizer_latency = 2

		def testbench():
			for i in range(10):
				yield fifo.w_data.eq(i)
				yield fifo.w_en.eq(1)
				yield

				if (i - ff_syncronizer_latency) > 0:
					self.assertEqual((yield fifo.r_level), i - ff_syncronizer_latency)
				else:
					self.assertEqual((yield fifo.r_level), 0)

		simulator = Simulator(fifo)
		simulator.add_clock(100e-6)
		simulator.add_sync_process(testbench)
		simulator.run()

	def check_async_fifo_level(self, fifo, fill_in, expected_level, read = False):
		write_done = Signal()

		def write_process():
			for i in range(fill_in):
				yield fifo.w_data.eq(i)
				yield fifo.w_en.eq(1)
				yield Tick('write')
			yield fifo.w_en.eq(0)
			yield Tick('write')
			yield Tick('write')
			self.assertEqual((yield fifo.w_level), expected_level)
			yield write_done.eq(1)

		def read_process():
			if read:
				yield fifo.r_en.eq(1)
			while not (yield write_done):
				yield Tick('read')
			self.assertEqual((yield fifo.r_level), expected_level)

		simulator = Simulator(fifo)
		simulator.add_clock(100e-6, domain = 'write')
		simulator.add_sync_process(write_process, domain = 'write')
		simulator.add_clock(50e-6, domain = 'read')
		simulator.add_sync_process(read_process, domain = 'read')
		with simulator.write_vcd('test.vcd'):
			simulator.run()

	def test_async_fifo_level(self):
		fifo = AsyncFIFO(width = 32, depth = 8, r_domain = 'read', w_domain = 'write')
		self.check_async_fifo_level(fifo, fill_in = 5, expected_level = 5)

	def test_async_fifo_level_full(self):
		fifo = AsyncFIFO(width = 32, depth = 8, r_domain = 'read', w_domain = 'write')
		self.check_async_fifo_level(fifo, fill_in = 10, expected_level = 8)

	def test_async_buffered_fifo_level(self):
		fifo = AsyncFIFOBuffered(width = 32, depth = 9, r_domain = 'read', w_domain = 'write')
		self.check_async_fifo_level(fifo, fill_in = 5, expected_level = 5)

	def test_async_buffered_fifo_level_only_three(self):
		fifo = AsyncFIFOBuffered(width = 32, depth = 9, r_domain = 'read', w_domain = 'write')
		self.check_async_fifo_level(fifo, fill_in = 3, expected_level = 3)

	def test_async_buffered_fifo_level_full(self):
		fifo = AsyncFIFOBuffered(width = 32, depth = 9, r_domain = 'read', w_domain = 'write')
		self.check_async_fifo_level(fifo, fill_in = 10, expected_level = 9)

	def test_async_buffered_fifo_level_empty(self):
		fifo = AsyncFIFOBuffered(width = 32, depth = 9, r_domain = 'read', w_domain = 'write')
		self.check_async_fifo_level(fifo, fill_in = 0, expected_level = 0, read = True)



# SPDX-License-Identifier: BSD-2-Clause

import sys
import subprocess
from pathlib import Path

from .utils  import ToriiTestSuiteCase

EXAMPLE_DIR = (Path(__file__).parent / '..' / 'examples').resolve()
class ExamplesTestCase(ToriiTestSuiteCase):

	def _example_test(name: str):
		example_path = (EXAMPLE_DIR / name).resolve()

		def test_example(self):
			subprocess.check_call([
					sys.executable, str(example_path),
					'generate'
				],
				stdout = subprocess.DEVNULL
			)
		return test_example

	test_alu      = _example_test('basic/alu.py')
	test_alu_hier = _example_test('basic/alu_hier.py')
	test_arst     = _example_test('basic/arst.py')
	test_cdc      = _example_test('basic/cdc.py')
	test_ctr      = _example_test('basic/ctr.py')
	test_ctr_en   = _example_test('basic/ctr_en.py')
	test_fsm      = _example_test('basic/fsm.py')
	test_gpio     = _example_test('basic/gpio.py')
	test_inst     = _example_test('basic/inst.py')
	test_mem      = _example_test('basic/mem.py')
	test_pmux     = _example_test('basic/pmux.py')
	test_por      = _example_test('basic/por.py')
	test_uart     = _example_test('basic/uart.py')
